/***

lrdirent.c -- Local Resources, directory tracking

***/

#define LR_NO_DIRENT_TRACKING 1
#include "lrdirent.h" /* pick up function prototypes */
#include "lrinit.h"   /* pick up EHRESTRK declaration */
#include "lrerrs.h"   /* local error codes */

/***
_lrOpendir()
Track opendir() calls.
***/

DIR* /* null if directory could not be opened, otherwise a valid DIR* */
_lrOpendir(               /* track opendir() calls */
   const char *filename,  /* file from which this function was called */
   const char *funcname,  /* function from which this function was called */
   int line,              /* line on which this function was called */
   const char *dirname)   /* name of directory to open */
{
   DIR *volatile pDir = NULL;  /* initialize to a non-valid value */

   EHCATCH(_lrOpendir) {       /* begin function body proper of catcher */

      pDir = opendir(dirname); /* attempt to obtain resource */
      if (pDir == NULL)        /* if unsuccessful, print error */
      {
         ehPError(filename, funcname, line, EH_IS_ERROR, LR_OPENDIR);
         break;                /* out to section three */
      }

      ehResObt(filename,funcname,line,&lrResTrk,
            (int)LR_RES_DIRENT, dirname,
            (pfv_ulv)closedir, 1L, 
            (ulong)pDir);

   } EHENDCATCH;

   /* recovery */
   if (ehError() && pDir)
   {
      closedir(pDir);
      pDir = NULL;
   }

   EHRETURN(pDir); /* optional for C++, required for C */
} /* _lrOpendir */


/***
_lrClosedir()
Track closedir() calls.
***/

void
_lrClosedir(const char *filename,
   const char *funcname,
   int line,
   DIR*pDir)
{
   EHCATCH(_lrClosedir) {
   /***
      no body but we need clean up section
      so that we are protected from exceptions
      being thrown...exception throwing from
      within a clean-up section is inhibited.
   ***/
   } EHENDCATCH;

   /* mark the resource and release it */
   ehResRel(filename,funcname,line,&lrResTrk,
         (int)LR_RES_DIRENT, (pfv_ulv)closedir, 1L, 
         (ulong)pDir);

   EHLEAVE(); /* optional for C++, required for C */
} /* _lrClosedir */


