/***

lrfildes.c -- Local Resources, file descriptors

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char*rcsid =
       "$Id: lrfildes.c,v 1.7 1996/03/08 07:35:10 ccobb Exp $";

#define LR_NO_FILDES_TRACKING 1
#include "lrfildes.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "lrinit.h"
#include "lrerrs.h"

EHFILENAME(lrfildes.c);

/***
_lrCreat()
Track creat() calls.
***/

int
_lrCreat(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   const char *pcPath,
   mode_t mode)

{
   volatile int iFilDes = -1;
   volatile bool completed = false;

   EHCATCH(_lrCreat) {

      iFilDes = creat(pcPath, mode);
      if (iFilDes == -1)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, LR_CREAT);
         break;
      }

      completed = ehResObt(pcFile,pcFunc,iLine,&lrResTrk,
            (int)LR_RES_FILDES, pcPath,
            (pfv_ulv)close, 1L, 
            (ulong)iFilDes);

   } EHENDCATCH;

   /* recovery */
   if (ehError() && !completed)
   {
      if (iFilDes != -1)
      {
         close(iFilDes);
         iFilDes = -1;
      }
   }

   EHRETURN(iFilDes);
} /* _lrCreat */


/***
_lrOpen()
Track open() calls.
***/

int
_lrOpen(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   const char *pcPath,
   int iOFlag)
{
   volatile int iFilDes = -1;
   volatile bool completed = false;

   EHCATCH(_lrOpen) {

      iFilDes = open(pcPath, iOFlag);
      if (iFilDes == -1)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, LR_OPEN);
         break;
      }

      completed = ehResObt(pcFile,pcFunc,iLine,&lrResTrk,
            (int)LR_RES_FILDES, pcPath,
            (pfv_ulv)close, 1L, 
            (ulong)iFilDes);

   } EHENDCATCH;

   /* recovery */
   if (ehError() && !completed)
   {
      if (iFilDes != -1)
      {
         close(iFilDes);
         iFilDes = -1;
      }
   }

   EHRETURN(iFilDes);
} /* _lrOpen */


/***
_lrDup()
Track dup() calls.
***/

int
_lrDup(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   int iFilDesOrg)
{
   volatile int iFilDesNew = -1;
   volatile bool completed = false;

   EHCATCH(_lrDup) {

      iFilDesNew = dup(iFilDesOrg);
      if (iFilDesNew == -1)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, LR_DUP);
         break;
      }

      completed = ehResObt(pcFile,pcFunc,iLine,&lrResTrk,
            (int)LR_RES_FILDES, "duped file descriptor",
            (pfv_ulv)close, 1L, 
            (ulong)iFilDesNew);

   } EHENDCATCH;

   /* recovery */
   if (ehError() && !completed)
   {
      if (iFilDesNew != -1)
      {
         close(iFilDesNew);
         iFilDesNew = -1;
      }
   }

   EHRETURN(iFilDesNew);
} /* _lrDup */


/***
_lrDup2()
Track dup2() calls.
***/

int 
_lrDup2(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   int iFilDes,  /* file descriptor to duplicate */
   int iFilDes2) /* file descriptor to close */
{
   volatile int iFilDesNew = -1;
   volatile bool completed = false;

   EHCATCH(_lrDup2)
   {

      /* attempt to dup2 the file descriptor */
      iFilDesNew = dup2(iFilDes, iFilDes2);
      if (iFilDesNew == -1)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_ERROR, LR_DUP2);
         break;
      }

      /***

         We need to check against stdin, stdout and stderr
         which might be reopened but are not already in 
         the resource tree.

         If the reopened file is stdin, stdout or stderr,
         then don't call ehResReObt().
         Call ehResObt() instead since we are not
         re-obtaining a resource that is in the resource
         tree.  However, if the same resource handle is
         returned (e.g., stderr), then we may have a
         problem because next time through, stderr will
         be a resource in the tree, but we won't know it.

      ***/

      if (iFilDes2 != 0 && iFilDes2 != 1 && iFilDes2 != 2)
      {
         /* mark the resource as obtained */
         completed = ehResReObt(pcFile,pcFunc,iLine,&lrResTrk,
               (int)LR_RES_FILDES, "dup2ed file descriptor",
               (ulong)iFilDes2,    /* resource that was released */
               (pfv_ulv)close, 1L, 
               (ulong)iFilDesNew);
      }

   } EHENDCATCH;

   /* recovery */
   if (ehError() && !completed)
   {
      if (iFilDesNew != -1 && iFilDesNew != iFilDes2)
      {
         close(iFilDesNew); /* close the new file */
         iFilDesNew = -1;
         /* now both files are closed */
      }
      /* mark the old file as closed */
      ehResRel(pcFile,pcFunc,iLine,&lrResTrk,
            (int)LR_RES_FILDES, (pfv_ulv)0, 1L,
            (ulong)iFilDes2);
   }
   FUN_RETURN(iFilDes2);
} /* _lrDup2 */


/***
_lrClose()
Track close() calls.
***/

void
_lrClose(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   int iFilDes)
{
   EHCATCH(_lrClose) {} EHENDCATCH; /* no body but we need a clean up section */

   if (iFilDes != 0 && iFilDes != 1 && iFilDes != 2)
   {
      /* mark the resource and release it */
      ehResRel(pcFile,pcFunc,iLine,&lrResTrk,
            (int)LR_RES_FILDES, (pfv_ulv)close, 1L, 
            (ulong)iFilDes);
   }

   EHLEAVE();
} /* _lrClose */


