/***

ehavltre.c -- Error Handler, AVL tree support for error handling library.

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char *rcsid =
       "$Id: ehavltre.c,v 1.3 1996/01/11 23:32:44 ccobb Exp $";

/* as_tree - tree library for as
 * vix 14dec85 [written]
 * vix 02feb86 [added tree balancing from wirth "a+ds=p" p. 220-221]
 * vix 06feb86 [added ehAvlMung()]
 * vix 20jun86 [added ehAvlDel per wirth a+ds (mod2 v.) p. 224]
 * vix 23jun86 [added delete uar to add for replaced nodes]
 * cjc 12jul93 [made ansi and eh compatible]
 */

/* This program text was created by Paul Vixie using examples from the
   book: "Algorithms & Data Structures," Niklaus Wirth, Prentice-Hall,
   1986, ISBN 0-13-022005-1.  This code and associated documentation
   is hereby placed in the public domain.
 */

#include "ehenv.h"

EHBEGINEXTERNC
#include <stdio.h>
#include <stdlib.h>
EHENDEXTERNC

#define NTRKMEM 1

#define EH_CHAN 2
#include "eh.h"
#include "eherrs.h"

#include "ehavltre.h"

EHFILENAME(ehavltre.c);

/* function prototypes */

static void avlAdd A((EHAVLNODE **volatile ppEhAvlNode, 
                      volatile pfi_pvpv pfiCompare, 
                      const void *volatile pv_data,
                      int *piReqBal, 
                      int *volatile piAdded));
static void avlAddNew A((EHAVLNODE **volatile ppEhAvlNode, 
                          const void *pvUserData,
                          int *piReqBal, 
                          int *volatile piAdded));
static void avlAddUp A((EHAVLNODE **ppEhAvlNode, 
                           pfi_pvpv pfiCompare, 
                           const void *pv_data,
                           int *piReqBal, 
                           int *piAdded));
static void avlAddDn A((EHAVLNODE **ppEhAvlNode, 
                            pfi_pvpv pfiCompare, 
                            const void *pv_data,
                            int *piReqBal, 
                            int *piAdded));

static int  avlDel    A((EHAVLNODE **ppEhAvlNode,
                         pfi_pvpv pfiCompare,
                         pfi_ppv  pfiDelete,
                         const void *pvUserData,
                         int *piReqBal,
                         int *pi_uar_called));
static void avlBalBtmHvy A((EHAVLNODE **ppEhAvlNode, int *piReqBal));
static void avlBalTopHvy A((EHAVLNODE **ppEhAvlNode, int *piReqBal));
static void avlDelNode   A((EHAVLNODE **ppEhAvlNodeNbr, int *piReqBal,
                            EHAVLNODE **ppEhAvlNodeQ, pfi_ppv pfiDelete,
                            int *pi_uar_called));
static void avlFreeNode   A((EHAVLNODE**ppEhAvlNode));
static void avlTrav       A((EHAVLNODE**ppEhAvlNode,
                           pfv_ppEhAvlNode pfiNodeFunc, pfi_ppv pfiUserFunc));
static const void*avlSearch A((EHAVLNODE**ppEhAvlNode,pfi_pvpv pfiCompare,
	                     const void*pvUserData));
static void avlTreeDump   A((EHAVLNODE *pEhAvlNode, pfv_pFilepv, int iLevel));
static void avlDumpNode   A((EHAVLNODE *pEhAvlNode,int iLevel));
static void avlUserDump   A((EHAVLNODE *pEhAvlNode,
      pfv_pFilepv pfvDmpUserData, FILE*));

EHAVL *
ehAvlInit A3(            /* allocate an initilized AVL tree structure */
   pfi_pvpv,    pfiCompare, /* user-supplied funct used for node comparisons */
   pfi_ppv,     pfiDelete,  /* user-supplied funct to delete user data */
   pfv_pFilepv, pfvDump)    /* user-supplied funct to dump user data */
{
   EHAVL*pEhAvl = NULL;

   // puts("ehAvlInit: about to call EHTRACE(ehAvlInit)");

   EHTRACE(ehAvlInit);

   // puts("ehAvlInit: back from EHTRACE(ehAvlInit)");

   pEhAvl = NEW(EHAVL);
   if (pEhAvl == NULL) EHTHROW_RETURN(EH_NEWMACRO, (EHAVL*)NULL);

   pEhAvl->pfiCompare = pfiCompare;
   pEhAvl->pfiDelete  = pfiDelete;
   pEhAvl->pfvDump    = pfvDump;

   // puts("ehAvlInit: leaving ehAvlInit()");

   FUN_RETURN(pEhAvl);
}  /* ehAvlInit */

/***
This function is called to terminate use of the 
AVL tree.  It tis the complement of ehAvlInit().
The user-supplied Delete function, if not null,
is applied to each pointer to user data.  Typically,
this would free or 'de-construct' the user data.
***/
void
ehAvlTerm A1(       /* terminate use of the avl tree */
   EHAVL *, pEhAvl) /* avl tree to terminate */
{

   pfv_ppEhAvlNode pfvAvlFreeNode = avlFreeNode; /* bug in sc */

   EHTRACE(ehAvlTerm);

   if (pEhAvl)
   {
      if (pEhAvl->pEhAvlNodeTop)
         avlTrav(&pEhAvl->pEhAvlNodeTop, pfvAvlFreeNode, pEhAvl->pfiDelete);
      EHFREE(pEhAvl);
   }
   FUN_LEAVE();
}  /* ehAvlTerm */


/***
avlTrav()
This function traverses the entire tree in order (top to bottom).
For each tree in the node, it applies the user-supplied function 
represented by pfiDelete.
This function is not used anywhere else in the avl code.
***/

static
void
avlTrav A3(
   EHAVLNODE**,     ppEhAvlNode,
   pfv_ppEhAvlNode, pfvAvlNodeFunc,
   pfi_ppv,         pfiUserFunc)
{
   EHTRACE(avlTrav);

   /* traverse to the top */
   if ((*ppEhAvlNode)->pEhAvlNodeUp)
   {
      avlTrav(&(*ppEhAvlNode)->pEhAvlNodeUp, pfvAvlNodeFunc, pfiUserFunc);
   }

   /* apply user function */
   if (pfiUserFunc)
   {
      (*pfiUserFunc)(&(*ppEhAvlNode)->pvUserData);
   }

   /* traverse to the bottom */
   if ((*ppEhAvlNode)->pEhAvlNodeDn)
   {
      avlTrav(&(*ppEhAvlNode)->pEhAvlNodeDn, pfvAvlNodeFunc, pfiUserFunc);
   }

   /* apply node function */
   if (pfvAvlNodeFunc)
   {
      (*pfvAvlNodeFunc)(ppEhAvlNode);
   }

   FUN_LEAVE();
} /* avlTrav() */


/***
This function de-allocates a node
and sets the node pointer to null
***/
static
void
avlFreeNode A1(
   EHAVLNODE **, ppEhAvlNode)
{
   EHTRACE(avlFreeNode);
   EHFREE(*ppEhAvlNode);
   FUN_LEAVE();
}  /* avlFreeNode */


/***
This function attemps to locate a tree node which contains user data that
matches the supplied user data.
***/
void *
ehAvlSrch A2(                /* search for a matching node in the tree */
   EHAVL *,      pEhAvl,     /* tree to search */
   const void *, pvUserData) /* user data used as a matching template */
{
   EHTRACE(ehAvlSrch);

   if (pEhAvl->pEhAvlNodeTop)
   {
      void *pv = (void*)avlSearch(&pEhAvl->pEhAvlNodeTop,
                            pEhAvl->pfiCompare,
                            pvUserData);
      FUN_RETURN(pv);
   }
   FUN_RETURN((void*)NULL);
} /* ehAvlSrch() */


/***
This function attemps to locate a tree node which contains user data that
matches the supplied user data.
***/
static
const void *
avlSearch A3(                 /* search tree: guts */
   EHAVLNODE**,  ppEhAvlNode, /* address of tree node pointer */
   pfi_pvpv,     pfiCompare,  /* compare function */
   const void *, pvUserData)  /* user data used as template */
{
   int iCmp;
   const void *pv;

   EHTRACE(avlSearch);

   /* check the current node */
   iCmp = (*pfiCompare)(pvUserData, (*ppEhAvlNode)->pvUserData);
   if (iCmp == 0) /* found it! */
   {
      FUN_RETURN( (**ppEhAvlNode).pvUserData );
   }

   pv = (void*)NULL;

   if (iCmp > 0)
   {
      /* check the lower tree */
      if ( (*ppEhAvlNode)->pEhAvlNodeDn )
      {
         pv = avlSearch(&(*ppEhAvlNode)->pEhAvlNodeDn, pfiCompare, pvUserData);
      }
   }
   else if (iCmp < 0)
   {
      /* check the upper tree */
      if ( (*ppEhAvlNode)->pEhAvlNodeUp )
      {
         pv = avlSearch(&(*ppEhAvlNode)->pEhAvlNodeUp, pfiCompare, pvUserData);
      }
   }
   FUN_RETURN(pv);
} /* avlSearch() */


/***
ehAvlAdd()
This function adds a new node to the  avl tree.  The user data is
attached to the new node.  If an existing node is found which exaclty
matches the new user data, the old user data is deleted and the new data
is attached.
***/

bool
ehAvlAdd A3(                /* add new node to the tree */
   EHAVL *,     pEhAvl,     /* tree to add node to */
   const void*, pvUserData, /* user data to attach to node */
   pfi_pvpv,    pfiCompare) /* optional compare function */
{              
   int   iBalance = false;
   int   iAdded = false; /* indicates if a new node has been added */

   EHTRACE(ehAvlAdd);

   EHASSERT_RETURN(pEhAvl != NULL, false);

   avlAdd(&(pEhAvl->pEhAvlNodeTop),
           pfiCompare ? pfiCompare : pEhAvl->pfiCompare,
           pvUserData,
           &iBalance,
           &iAdded);

   FUN_RETURN(true);
} /* ehAvlAdd() */

/***
avlAdd()
This function adds a new node to the tree.
If an existing node already exists having the value (according
to the supplied compare function) the old node will be discarded
(using the supplied delete function) and then replaced with the new
node supplied.  If no matching node exists, a new tree node will be 
added and the user data attached.
avlAdd() will fail on lack of memory of if either the user-supplied
compared or delete function fail (perhaps due to an interrupt).
***/

static 
void
avlAdd A5(                            /* guts: add new node to tree */
   EHAVLNODE **volatile, ppEhAvlNode, /* address of tree pointer */
   volatile pfi_pvpv,    pfiCompare,  /* compare function */
   const void *volatile, pvUserData,  /* user data to attach to node */
   int *,                piReqBal,   /* does tree need to be balanced? */
   int *volatile,        piAdded)     /* has node been added? */
{                            
   int  cmp = 0;

   EHTRACE(avlAdd);

   /***
      Is there a node here?
      If NOT, add a node,
      set the balance flag, then exit.
   ***/
   if (*ppEhAvlNode == NULL)
   {
      avlAddNew(ppEhAvlNode, pvUserData, piReqBal, piAdded);
      FUN_LEAVE();
   }

   /***
      There IS a node here:
      compare the user-supplied data
      with the current data in sub-tree
      using the user-supplied compare function
   ***/
   cmp = (*pfiCompare)(pvUserData, (*ppEhAvlNode)->pvUserData);

   /*** if LESS, move up. ***/
   if (cmp < 0)
   {
      avlAddUp(ppEhAvlNode, pfiCompare, pvUserData, piReqBal, piAdded);
      FUN_LEAVE();
   }

   /*** if MORE, move down ***/
   if (cmp > 0)
   {
      avlAddDn(ppEhAvlNode, pfiCompare, pvUserData, piReqBal, piAdded);
      FUN_LEAVE();
   }

   /*** not less, not more: this is the same key!  error... ***/
   EHTHROW_LEAVE(EH_AVL_DUP);
} /* avlAdd() */


/***
avlAddNew()
This function adds a new node to the tree
and attaches to user data.
***/

static 
void
avlAddNew A4(                        /* allocate the new node for the tree */
   EHAVLNODE**volatile, ppEhAvlNode, /* address of tree pointer */
   const void *,        pvUserData,  /* user data to attach to node */
   int *,               piReqBal,   /* does tree need balancing? */
   int *volatile,       piAdded)     /* has node been added? */
{
   EHCATCH(avlAddNew) {

      EHDBG("no node here: adding new node, setting h=true");

      if ((*ppEhAvlNode = NEW(EHAVLNODE)) == NULL) EHTHROW(EH_NEWMACRO);

      (*ppEhAvlNode)->pEhAvlNodeSelf = *ppEhAvlNode; /* self-referent address */
      (*ppEhAvlNode)->pEhAvlNodeUp = NULL;
      (*ppEhAvlNode)->pEhAvlNodeDn = NULL;
      (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
      (*ppEhAvlNode)->pvUserData = pvUserData;
      *piReqBal = true;
      *piAdded = true;

   } EHENDCATCH;

   if (ehError())
   {
      *piAdded = false;
      if (*ppEhAvlNode) free(*(EHAVLNODE**)ppEhAvlNode);
      *ppEhAvlNode = NULL;
   }
   FUN_LEAVE();
} /* avlAddNew() */


/***
avlAddUp()
This function attemps to add a new node to the upper subtree 
and balances the tree if a node was successfully added.
***/

static 
void
avlAddUp A5(
   EHAVLNODE **,  ppEhAvlNode,
   pfi_pvpv,      pfiCompare,
   const void *,  pvUserData,
   int *,         piReqBal,
   int *,         piAdded)
{
   EHAVLNODE   *p1, *p2;

   EHTRACE(avlAddUp);

   EHDBG("LESS. adding up.");

   /* now, add a node here */
   avlAdd(&(*ppEhAvlNode)->pEhAvlNodeUp, pfiCompare, pvUserData, piReqBal,
         piAdded);

   if (*piReqBal)
   {   /* upper branch has grown longer */
      EHDBG("LESS: upper branch has grown");
      switch ((*ppEhAvlNode)->EhBalStat)
      {
      case EH_BTM_HVY:   /* lower branch WAS longer; balance is ok now */
         EHDBG("LESS: case EH_BTM_HVY.. balance restored implicitly");
         (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
         *piReqBal = false;
         break;
      case EH_BALANCED:   /* balance WAS okay; now upper branch longer */
         EHDBG("LESS: case EH_BALANCED.. balance bad but still ok");
         (*ppEhAvlNode)->EhBalStat = EH_TOP_HVY;
         break;
      case EH_TOP_HVY:
         /* upper branch was already too long. rebalance */
         EHDBG("LESS: case EH_TOP_HVY: rebalancing");
         p1 = (*ppEhAvlNode)->pEhAvlNodeUp;
         if (p1->EhBalStat == EH_TOP_HVY)
         {   /* LL */
            EHDBG("LESS: single LL");
            (*ppEhAvlNode)->pEhAvlNodeUp = p1->pEhAvlNodeDn;
            p1->pEhAvlNodeDn = *ppEhAvlNode;
            (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
            *ppEhAvlNode = p1;
         }
         else
         {         /* double LR */
            EHDBG("LESS: double LR");

            p2 = p1->pEhAvlNodeDn;
            p1->pEhAvlNodeDn = p2->pEhAvlNodeUp;
            p2->pEhAvlNodeUp = p1;

            (*ppEhAvlNode)->pEhAvlNodeUp = p2->pEhAvlNodeDn;
            p2->pEhAvlNodeDn = *ppEhAvlNode;

            (*ppEhAvlNode)->EhBalStat = ((p2->EhBalStat == EH_TOP_HVY) ?
                  EH_BTM_HVY : EH_BALANCED);
            p1->EhBalStat =             ((p2->EhBalStat == EH_BTM_HVY) ?
                  EH_TOP_HVY : EH_BALANCED);
            *ppEhAvlNode = p2;
         }
         (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
         *piReqBal = false;
      } /*switch*/
   } /*if*/

   FUN_LEAVE();
} /* avlAddUp() */


/***
avlAddDn()
This function attemps to add a new node to the lower subtree 
and balances the tree if a node was successfully added.
***/

static 
void
avlAddDn A5(
   EHAVLNODE **, ppEhAvlNode,
   pfi_pvpv,     pfiCompare,
   const void *, pvUserData,
   int *,        piReqBal,
   int *,        piAdded)
{
   EHAVLNODE   *p1, *p2;

   EHTRACE(avlAddDn);

   EHDBG("MORE: sprouting to the bottom");

   avlAdd(&(*ppEhAvlNode)->pEhAvlNodeDn, pfiCompare, pvUserData, piReqBal,
      piAdded);

   if (*piReqBal)
   {   /* lower branch has grown longer */
      EHDBG("MORE: lower branch has grown");
      switch ((*ppEhAvlNode)->EhBalStat)
      {
      case EH_TOP_HVY:
         EHDBG("MORE: balance was off, fixed implicitly");
         (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
         *piReqBal = false;
         break;
      case EH_BALANCED:   
         EHDBG("MORE: balance was okay, now off but ok");
         (*ppEhAvlNode)->EhBalStat = EH_BTM_HVY;
         break;
      case EH_BTM_HVY:   
         EHDBG("MORE: balance was off, need to rebalance");
         p1 = (*ppEhAvlNode)->pEhAvlNodeDn;
         if (p1->EhBalStat == EH_BTM_HVY)
         {   /* RR */
            EHDBG("MORE: single RR");
            (*ppEhAvlNode)->pEhAvlNodeDn = p1->pEhAvlNodeUp;
            p1->pEhAvlNodeUp = *ppEhAvlNode;
            (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
            *ppEhAvlNode = p1;
         }
         else
         {         /* double RL */
            EHDBG("MORE: double RL");

            p2 = p1->pEhAvlNodeUp;
            p1->pEhAvlNodeUp = p2->pEhAvlNodeDn;
            p2->pEhAvlNodeDn = p1;

            (*ppEhAvlNode)->pEhAvlNodeDn = p2->pEhAvlNodeUp;
            p2->pEhAvlNodeUp = *ppEhAvlNode;

            (*ppEhAvlNode)->EhBalStat = ((p2->EhBalStat == EH_BTM_HVY) ?
                  EH_TOP_HVY : EH_BALANCED);
            p1->EhBalStat =             ((p2->EhBalStat == EH_TOP_HVY) ?
                  EH_BTM_HVY : EH_BALANCED);
            *ppEhAvlNode = p2;
         }
         (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
         *piReqBal = false;
      } /*switch*/
   }

   FUN_LEAVE();
} /* avlAddDn() */


/***
ehAvlDel()
This function attempts to locate and delete the node 
which matches the user data supplied.
This is the public entry point for deleting a node.
***/

int
ehAvlDel A3(                 /* delete node from tree */
   EHAVL*,       pEhAvl,     /* tree to delete from */
   const void *, pvUserData, /* user data as template for match */
   pfi_pvpv,    pfiCompare)  /* optional compare function */
{
   int iBalance = false;
   int i_uar_called = false;
   volatile int iRetVal = 0;

   EHTRACE(ehAvlDel);

   if (pEhAvl->pEhAvlNodeTop)
   {
      iRetVal = avlDel(&(pEhAvl->pEhAvlNodeTop),
                         pfiCompare ? pfiCompare : pEhAvl->pfiCompare,
                         pEhAvl->pfiDelete,
                         pvUserData,
                         &iBalance,
                         &i_uar_called);
   }

   FUN_RETURN(iRetVal);
}  /* ehAvlDel */


/***
avlDel()
This function is the recursive entry point for deleting a node.
***/

static 
int
avlDel A6(                   /* perform node deletion dirty work */
   EHAVLNODE **, ppEhAvlNode,   /* address of tree pointer */
   pfi_pvpv,     pfiCompare,    /* compare function */
   pfi_ppv,      pfiDelete,    /* user data deletion function */
   const void *, pvUserData,    /* user data used as match template */
   int *,        piReqBal,     /* tree needs balancing */
   int *,        pi_uar_called) /* ??? */
{
   EHAVLNODE   *pEhAvlNodeQ;
   int   iComp;
   volatile int iRetVal = 0;

   EHTRACE(avlDel);

   /* find the node to delete */
   iComp = (*pfiCompare)(pvUserData, (*ppEhAvlNode)->pvUserData);
   if (iComp < 0) /* the node we're looking for is above us */
   {
      EHDBG("too high - scan up");
      if ((*ppEhAvlNode)->pEhAvlNodeUp == NULL)
         FUN_RETURN(-1);
      iRetVal = avlDel(&(*ppEhAvlNode)->pEhAvlNodeUp,
                       pfiCompare,
                       pfiDelete,
                       pvUserData,
                       piReqBal,
                       pi_uar_called);
      if (*piReqBal)
      {
         avlBalBtmHvy(ppEhAvlNode, piReqBal);
      }
   }
   else if (iComp > 0) /* the node we're looking for is below us */
   {
      EHDBG("too low - scan down");
      if ((*ppEhAvlNode)->pEhAvlNodeDn == NULL)
         FUN_RETURN(-1);
      iRetVal = avlDel(&(*ppEhAvlNode)->pEhAvlNodeDn,
                       pfiCompare, 
                       pfiDelete,
                       pvUserData,
                       piReqBal,
                       pi_uar_called);
      if (*piReqBal)
      {
         avlBalTopHvy(ppEhAvlNode, piReqBal);
      }
   }
   else /* delete THIS node ! */
   {
      EHDBG("equal");
      pEhAvlNodeQ = *ppEhAvlNode; /* the matching node */
      if (pEhAvlNodeQ->pEhAvlNodeUp == NULL)
      {
         EHDBG("upper subtree null");
         *ppEhAvlNode = pEhAvlNodeQ->pEhAvlNodeDn; /* pnt par to lower node */
         *piReqBal = true;
      }
      else if (pEhAvlNodeQ->pEhAvlNodeDn == NULL)
      {
         EHDBG("lower subtree null");
         *ppEhAvlNode = pEhAvlNodeQ->pEhAvlNodeUp; /* pnt par to upper node */
         *piReqBal = true;
      }
      else /* yikes! both sub-trees are populated */
      {
         EHDBG("neither subtree null");
         avlDelNode(&pEhAvlNodeQ->pEhAvlNodeUp, piReqBal, &pEhAvlNodeQ,
               pfiDelete, pi_uar_called);
         if (*piReqBal)
         {
            avlBalBtmHvy(ppEhAvlNode, piReqBal);
         }
      }
      if (!*pi_uar_called && pfiDelete)
      {
         (*pfiDelete)(&pEhAvlNodeQ->pvUserData);
         pEhAvlNodeQ->pvUserData = NULL;
      }
      free(pEhAvlNodeQ);
   }

   FUN_RETURN(iRetVal);
} /* avlDel */


/***
avlDelNode()
This function recursively deletes a node when neither 
the upper nor lower subtrees are NULL.
***/

static 
void
avlDelNode A5(
   EHAVLNODE**, ppEhAvlNodeNbr,
   int*,        piReqBal,
   EHAVLNODE**, ppEhAvlNodeQ,
   pfi_ppv,     pfiDelete,
   int*,        pi_uar_called)
{

   EHTRACE(avlDelNode);

   if ((*ppEhAvlNodeNbr)->pEhAvlNodeDn != NULL)
   {
      avlDelNode(&(*ppEhAvlNodeNbr)->pEhAvlNodeDn, piReqBal, ppEhAvlNodeQ,
            pfiDelete, pi_uar_called);
      if (*piReqBal)
      {
         avlBalTopHvy(ppEhAvlNodeNbr, piReqBal);
      }
   }
   else
   {
      if (pfiDelete)
      {
         (*pfiDelete)(&(*ppEhAvlNodeQ)->pvUserData);
         (*ppEhAvlNodeQ)->pvUserData = NULL;
      }
      *pi_uar_called = true;
      (*ppEhAvlNodeQ)->pvUserData = (*ppEhAvlNodeNbr)->pvUserData;
      *ppEhAvlNodeQ = *ppEhAvlNodeNbr;
      *ppEhAvlNodeNbr = (*ppEhAvlNodeNbr)->pEhAvlNodeUp;
      *piReqBal = true;
   }

   FUN_LEAVE();
} /* avlDelNode */


/***
avlBalBtmHvy()
This function is used within the context of deleting a node
above this node.
***/

static 
void
avlBalBtmHvy A2(
   EHAVLNODE**, ppEhAvlNode,
   int*,   piReqBal)
{
   EHAVLNODE   *pEhAvlNodeHvy;
   EHBALSTAT   EhBalStatHvy;

   EHTRACE(avlBalBtmHvy);

   EHDBG("upper branch has shrunk");

   switch ((*ppEhAvlNode)->EhBalStat)
   {
   case EH_TOP_HVY: 
      EHDBG("was imbalanced, fixed implicitly");
      (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
      break;

   case EH_BALANCED:   
      EHDBG("was okay, is now one off");
      (*ppEhAvlNode)->EhBalStat = EH_BTM_HVY;
      *piReqBal = false;
      break;

   case EH_BTM_HVY:   EHDBG("was already off, this is too much");
      pEhAvlNodeHvy = (*ppEhAvlNode)->pEhAvlNodeDn;
      EhBalStatHvy = pEhAvlNodeHvy->EhBalStat;
      if (EhBalStatHvy != EH_TOP_HVY)
      {
         EHDBG("single RR");
         (*ppEhAvlNode)->pEhAvlNodeDn = pEhAvlNodeHvy->pEhAvlNodeUp;
         pEhAvlNodeHvy->pEhAvlNodeUp = *ppEhAvlNode;
         if (EhBalStatHvy == EH_BALANCED)
         {
            EHDBG("EhBalStatHvy == EH_BALANCED");
            (*ppEhAvlNode)->EhBalStat = EH_BTM_HVY;
            pEhAvlNodeHvy->EhBalStat = EH_TOP_HVY;
            *piReqBal = false;
         }
         else
         {
            EHDBG("EhBalStatHvy != EH_BALANCED");
            (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
            pEhAvlNodeHvy->EhBalStat = EH_BALANCED;
         }
         *ppEhAvlNode = pEhAvlNodeHvy;
      }
      else /* EhBalStatHvy is EH_TOP_HVY */
      {
         EHAVLNODE   *pEhAvlNodeHvyHvy;
         EHBALSTAT   EhBalStatHvyHvy;

         EHDBG("double RL");
         pEhAvlNodeHvyHvy = pEhAvlNodeHvy->pEhAvlNodeUp;
         EhBalStatHvyHvy = pEhAvlNodeHvyHvy->EhBalStat;
         pEhAvlNodeHvy->pEhAvlNodeUp = pEhAvlNodeHvyHvy->pEhAvlNodeDn;
         pEhAvlNodeHvyHvy->pEhAvlNodeDn = pEhAvlNodeHvy;
         (*ppEhAvlNode)->pEhAvlNodeDn = pEhAvlNodeHvyHvy->pEhAvlNodeUp;
         pEhAvlNodeHvyHvy->pEhAvlNodeUp = *ppEhAvlNode;
         (*ppEhAvlNode)->EhBalStat = ((EhBalStatHvyHvy == EH_BTM_HVY) ?
               EH_TOP_HVY : EH_BALANCED);
         pEhAvlNodeHvy->EhBalStat =  ((EhBalStatHvyHvy == EH_TOP_HVY) ?
               EH_BTM_HVY : EH_BALANCED);
         *ppEhAvlNode = pEhAvlNodeHvyHvy;
         pEhAvlNodeHvyHvy->EhBalStat = EH_BALANCED;
      }
   }

   FUN_LEAVE();
} /* avlBalBtmHvy */


/***
avlBalTopHvy()
This function is used within the context of deleting a node
below this node.
***/

static 
void
avlBalTopHvy A2(
   EHAVLNODE**, ppEhAvlNode,
   int*,   piReqBal)
{
   EHAVLNODE   *pEhAvlNodeHvy;
   EHBALSTAT   EhBalStatHvy;

   EHTRACE(avlBalTopHvy);

   EHDBG("lower branch has shrunk");

   switch ((*ppEhAvlNode)->EhBalStat)
   {
   case EH_BTM_HVY:   
      EHDBG("was imbalanced, fixed implicitly");
      (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
      break;

   case EH_BALANCED:   
      EHDBG("was okay, is now one off");
      (*ppEhAvlNode)->EhBalStat = EH_TOP_HVY;
      *piReqBal = false;
      break;

   case EH_TOP_HVY: 
      EHDBG("was already off, this is too much");
      pEhAvlNodeHvy = (*ppEhAvlNode)->pEhAvlNodeUp;
      EhBalStatHvy = pEhAvlNodeHvy->EhBalStat;
      if (EhBalStatHvy != EH_BTM_HVY)
      {
         EHDBG("single LL");
         (*ppEhAvlNode)->pEhAvlNodeUp = pEhAvlNodeHvy->pEhAvlNodeDn;
         pEhAvlNodeHvy->pEhAvlNodeDn = *ppEhAvlNode;
         if (EhBalStatHvy == EH_BALANCED)
         {
            EHDBG("EhBalStatHvy == EH_BALANCED");
            (*ppEhAvlNode)->EhBalStat = EH_TOP_HVY;
            pEhAvlNodeHvy->EhBalStat = EH_BTM_HVY;
            *piReqBal = false;
         }
         else
         {
            EHDBG("EhBalStatHvy != EH_BALANCED");
            (*ppEhAvlNode)->EhBalStat = EH_BALANCED;
            pEhAvlNodeHvy->EhBalStat = EH_BALANCED;
         }
         *ppEhAvlNode = pEhAvlNodeHvy;
      }
      else
      {
         EHAVLNODE   *pEhAvlNodeHvyHvy;
         EHBALSTAT   EhBalStatHvyHvy;

         EHDBG("double LR");
         pEhAvlNodeHvyHvy = pEhAvlNodeHvy->pEhAvlNodeDn;
         EhBalStatHvyHvy = pEhAvlNodeHvyHvy->EhBalStat;
         pEhAvlNodeHvy->pEhAvlNodeDn = pEhAvlNodeHvyHvy->pEhAvlNodeUp;
         pEhAvlNodeHvyHvy->pEhAvlNodeUp = pEhAvlNodeHvy;
         (*ppEhAvlNode)->pEhAvlNodeUp = pEhAvlNodeHvyHvy->pEhAvlNodeDn;
         pEhAvlNodeHvyHvy->pEhAvlNodeDn = *ppEhAvlNode;
         (*ppEhAvlNode)->EhBalStat = ((EhBalStatHvyHvy == EH_TOP_HVY) ?
               EH_BTM_HVY : EH_BALANCED);
         pEhAvlNodeHvy->EhBalStat =  ((EhBalStatHvyHvy == EH_BTM_HVY) ?
               EH_TOP_HVY : EH_BALANCED);
         *ppEhAvlNode = pEhAvlNodeHvyHvy;
         pEhAvlNodeHvyHvy->EhBalStat = EH_BALANCED;
      }
   }

   FUN_LEAVE();
} /* avlBalTopHvy */


/***
ehAvlTrav()
This function traverses the entire tree in order (top to bottom).
For each tree in the node, it applies the user-supplied function 
represented by pfiDelete.
This function is not used anywhere else in the avl code.
***/

int
ehAvlTrav A3(
   EHAVL*, pEhAvl,
   pfv_ppEhAvlNode, pfvAvlNodeFunc,
   pfi_ppv, pfiUserFunc)
{
   volatile int iRetVal = true;

   EHTRACE(ehAvlTrav);

   if (pEhAvl->pEhAvlNodeTop)
      avlTrav(&pEhAvl->pEhAvlNodeTop, pfvAvlNodeFunc, pfiUserFunc);

   FUN_RETURN(iRetVal);
} /* ehAvlTrav() */


/***
ehAvlTreeDump()
print out the members of the EHAVL structure.
***/

void
ehAvlTreeDump A1(
   EHAVL*, pEhAvl)
{
   fprintf(EHSTDERR, "------------- tree dump -------------\n");
   fflush(EHSTDERR);

   if (pEhAvl->pEhAvlNodeTop)
      avlTreeDump(pEhAvl->pEhAvlNodeTop,
              pEhAvl->pfvDump,
              0);

   fprintf(EHSTDERR, "-------------------------------------\n");
   fflush(EHSTDERR);
} /* ehAvlTreeDump */


/***
avlTreeDump()
print out the members of the EHAVL structure.
***/

static
void
avlTreeDump A3(
   EHAVLNODE*,  pEhAvlNode,
   pfv_pFilepv, pfvDumpUserData,
   int,         iLevel)
{
   if (pEhAvlNode->pEhAvlNodeUp)
      avlTreeDump(pEhAvlNode->pEhAvlNodeUp,
              pfvDumpUserData,
              iLevel+1);

   avlDumpNode(pEhAvlNode, iLevel+1);
   if (pfvDumpUserData)
   {
      fprintf(EHSTDERR, "%*s", (iLevel+1)*3, "");
      (*pfvDumpUserData)(EHSTDERR, pEhAvlNode->pvUserData);
      fprintf(EHSTDERR, "\n");
   }

   if (pEhAvlNode->pEhAvlNodeDn)
      avlTreeDump(pEhAvlNode->pEhAvlNodeDn,
              pfvDumpUserData,
              iLevel+1);

} /* avlTreeDump */


/***
avlDumpNode()
This function prints out the contents
of an AvlNode and the corresponding user data.
***/

static
void
avlDumpNode A2(
   EHAVLNODE*,    pEhAvlNode,
   int,       iLevel)
{
   fprintf(EHSTDERR, "%*s", iLevel*3, "");
   fprintf(EHSTDERR, "pEhAvlNodeUp: 0x%p\n", pEhAvlNode->pEhAvlNodeUp);

   if (iLevel == 1)
      fprintf(EHSTDERR, "==>");
   else
      fprintf(EHSTDERR, "%*s", iLevel*3, "");

   fprintf(EHSTDERR, " pEhAvlNode: 0x%p, pEhAvlNodeSelf: 0x%p\n",
         pEhAvlNode, pEhAvlNode->pEhAvlNodeSelf);

   fprintf(EHSTDERR, "%*s", iLevel*3, "");
   fprintf(EHSTDERR, "EhBalStat: %s, pvUserData: 0x%p\n", 
         pEhAvlNode->EhBalStat == EH_TOP_HVY ? "TOP" :
         pEhAvlNode->EhBalStat == EH_BTM_HVY ? "BTM" : "BAL",
         pEhAvlNode->pvUserData);

   fprintf(EHSTDERR, "%*s", iLevel*3, "");
   fprintf(EHSTDERR, "pEhAvlNodeDn: 0x%p\n", pEhAvlNode->pEhAvlNodeDn);

   fflush(EHSTDERR);
} /* avlDumpNode */


/***
ehAvlDump()
Dump sorted user data from tree using user-supplied dump routine.
***/

void
ehAvlDump A2(
   EHAVL*, pEhAvl,
   FILE*,  pFile)
{
   if (pEhAvl->pEhAvlNodeTop)
      avlUserDump(pEhAvl->pEhAvlNodeTop,
              pEhAvl->pfvDump, pFile);
   fprintf(pFile, "\n");
} /* ehAvlDump */


/***
avlUserDump()
This function dumps out the contents of 
an AVL Tree node.  Only a single node is printed.
***/

static
void
avlUserDump A3(
   EHAVLNODE*, pEhAvlNode, 
   pfv_pFilepv, pfvDumpUserData,
   FILE*,  pFile)
{
   if (pEhAvlNode->pEhAvlNodeUp)
      avlUserDump(pEhAvlNode->pEhAvlNodeUp,
              pfvDumpUserData, pFile);

   if (pfvDumpUserData)
   {
      (*pfvDumpUserData)(pFile, pEhAvlNode->pvUserData);
   }

   if (pEhAvlNode->pEhAvlNodeDn)
      avlUserDump(pEhAvlNode->pEhAvlNodeDn,
              pfvDumpUserData, pFile);

} /* avlUserDump() */


/***

$Log: ehavltre.c,v $
 * Revision 1.3  1996/01/11  23:32:44  ccobb
 * Fixed a memory leak in ehAvlTerm().
 * Added function ehResSetRelBy() to support AbortIO().
 *
 * Revision 1.2  1995/12/04  21:45:59  ccobb
 * Fully supports couplets.
 *
 * Revision 1.1  1995/11/27  06:04:56  ccobb
 * cleaned up and finished coverage testing on ehavltre.c
 *
 * Revision 1.3  1995/08/16  19:44:17  ccobb
 * Compiled with C++.
 *
 * Revision 1.2  1995/04/09  08:29:03  ccobb
 * Well, it works now!  I can't remember all the changes.
 *

***/
