/***

ehdivsrc.c -- Error Handler, divide error message source into two files
   o a header file containing #defines
   o an error-message text file containing error messages

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char *rcsid =
       "$Id: ehdivsrc.c,v 1.6 1996/04/12 21:54:59 ccobb Exp $";

#include "ehenv.h"

EHBEGINEXTERNC
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
EHENDEXTERNC

#include "eh.h"
#include "ehfuntrc.h"
#include "ehdivsrc.h"
#include "ehercode.h"

#ifndef isalpha
int     isalpha A((int c));
#endif

#define DEBUGGING 0

static void ehDivSrc A1(char *,pcFileID);

/***
ehDivSrc()
Divide error source file into header and error-text file.  
This function parses a text file which contains lines of the form:

SYMBOL = any text here that you like

The SYMBOL should uniquely identify each error message.  The error
messages should be placed to the right of the equals (=) sign.
All of the symbols to the left of the '=' will be collected 
and placed in a header file with the form:

#define SYMBOL 0x12345678,0x87654321 |* message text *|

where 
   o 0x12345678 is a 32-bit number representing
	 a time stamp to ensure compatability between source and message file (9 bits)
     the file offset to the message (23 bits)
   o 0x87654321 is a 32-bit number representing 
     the subsystem (16 bits) and
	 the error number (16 bits)

The first line of the error message text file an hex-ascii numeric string
representing the 9-bit time stamp.  This is checked when the file is opened.
Each subsequent line of the file has the format:

SYMBOL: message text

where
   SYMBOL is symbol from the .src file, and
   message text is the text from the file.

***/

static
void 
ehDivSrc A1(
   char*, pcFileID) /*{aaaa-z999} id for generated hdr and msg files */
{
   FILE *volatile pFilSrc = NULL; /* original "source" */
   FILE *volatile pFilHdr = NULL; /* header file */
   FILE *volatile pFilMsg = NULL; /* message file */
   volatile ER Er;                /* error-handler error */
   time_t ttTimeStamp;  /* time-stamp for file */
   ulong  ulTimeStamp;   /*  9-bit time stamp fragment */
   int    i;
   char   acSrcNam[EH_SZFILNAM+1];  /* name of original file */
   char   acHdrNam[EH_SZFILNAM+1];    /* name of header file to create */
   char   acMsgNam[EH_SZFILNAM+1];    /* name message text */
   char   acMsg[EH_MAXMSGLEN+1]; /* message file output buffer */
   char   acBuf[128]; /* read buffer */
   char   acHdr[128]; /* header file output buffer */

   // puts("ehDivSrc: about to call FUN_CATCH(ehDivSrc)");

   // FUN_CATCH("ehDivSrc");

   // ehSetOut(1);
   // ehSetFun(-1, 1);

   // puts("ehDivSrc: back from FUN_CATCH(ehDivSrc)");

   strcpy(Er.acPrefix, pcFileID);
   Er.hCode = 0;

   sprintf(acSrcNam, EH_SRC_FORMAT, Er.acPrefix);
   sprintf(acHdrNam, EH_HDR_FORMAT, Er.acPrefix);
   sprintf(acMsgNam, EH_MSG_FORMAT, Er.acPrefix);

   /* open original file for reading */
   if ((pFilSrc = fopen(acSrcNam, "r")) == NULL)
   {
      fprintf(stderr, "Cannot open %s\n", acSrcNam);
      fflush(stderr);
      goto Exit;
   }

   /* open header file for writing */
   if ((pFilHdr = fopen(acHdrNam, "w")) == NULL)
   {
      fprintf(stderr, "Cannot open %s\n", acHdrNam);
      fflush(stderr);
      goto Exit;
   }

   /* open message file for writing */
   if ((pFilMsg = fopen(acMsgNam, "w")) == NULL)
   {
      fprintf(stderr, "Cannot open %s\n", acMsgNam);
      fflush(stderr);
      goto Exit;
   }

   ttTimeStamp = time(NULL);
   ulTimeStamp = (ulong)ttTimeStamp << OFFSET_WIDTH;
   fprintf(pFilMsg, "0x%lx\n\n", ulTimeStamp);

   /* while there are lines to read */
   Er.hCode = EH_ERRMSG_STARTNUM;  /* # of 1st value to be generated */
   while (fgets(acBuf, (int)SZARR(acBuf), (FILE *)pFilSrc) != NULL)
   {
#if DEBUGGING
      printf("acBuf: [%s]\n", acBuf);
#endif

      /* put symbol into symbol var, message into message var */
      memset(acMsg, '\0', sizeof(acMsg));
      i = sscanf(acBuf, "%s = %[^\n]", acHdr, acMsg);

#if DEBUGGING
      printf("i: [%d]\n", i);
      printf("acHdr: [%s]\n", acHdr);
      printf("acMsg: [%s]\n", acMsg);
#endif

      assert(strlen(acMsg) < SZARR(acMsg));
#if ehDCC
      acMsg[strlen(acMsg) - 1] = '\0';
#endif
      if (i != 2 || *acBuf == '#')
      {
         continue;
      }
#if DEBUGGING
      printf("#define %s\t0x%08lxL,0x%08lxL\t/* %s */\n", 
            acHdr,  ulTimeStamp + ftell(pFilMsg), Er.lNo++, acMsg);
      printf("%s: %s\n", acHdr, acMsg);
#endif

      /*    write line to header file with associated value */
      fprintf((FILE *)pFilHdr, "#define %s\t0x%08lxL,0x%08lxL\t/* %s */\n", 
            acHdr,  ulTimeStamp + ftell(pFilMsg), ehHash(Er), acMsg);
      ++Er.hCode; /* portably increment error code */

      /*    write message to message file */
      fprintf((FILE *)pFilMsg, "%s: %s\n", acHdr, acMsg);
   }  /* end while */

Exit: /* Exit Sequence */

   // puts("ehDivSrc: in Exit");

   if (pFilMsg)
      fclose((FILE *)pFilMsg);

   if (pFilHdr)
      fclose((FILE *)pFilHdr);

   if (pFilSrc)
      fclose((FILE *)pFilSrc);

   // puts("ehDivSrc: leaving");

   // FUN_LEAVE();
} /* ehDivSrc */


int
main A2(int,    argc,
        char**, argv)
{
   int iRetVal = 0;

   /* usage: no parameters: error: give usage message */
   /*        1 parameter: parameter is input file name */
   /*                     parameter.h is header file */
   /*                     parameter.msg is msg file */

   // puts("main: entering");

   if (argc < 2)
   {
      printf(
         "\ndivsrc: divide error \"source\" file into .h and .msg files\n");
      printf("usage: divsrc <file id code>\n");
      iRetVal = 1;
      goto Exit;
   }

   // ehInitChnVol(EH_NEWDBGVAL);

   //ehSetOut(1);
   //ehSetFun(-1, 1);
   //ehSetVol(-1, 1);
   //ehSetMem(-1, 1);

   ehDivSrc(argv[1]);

   // puts("main: back from ehDivSrc()");

Exit:
   // puts("main: Exit section");

   // ehTermChnVol();

   // puts("main: leaving");

   return (iRetVal);
}


/***

$Log: ehdivsrc.c,v $
 * Revision 1.6  1996/04/12  21:54:59  ccobb
 * Now with xh support.
 *
 * Revision 1.4  1995/11/27  05:48:52  ccobb
 * compiled with sasc 6.56 patch
 *
 * Revision 1.3  1995/08/16  19:46:38  ccobb
 * Compiled with C++.
 *
 * Revision 1.2  1995/04/09  08:31:00  ccobb
 * Well, it works now!  I can't remember all the changes.
 *

***/
