/***

eherrlog.c -- Error Handler, error log file support

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char *rcsid =
       "$Id: eherrlog.c,v 1.11 1996/05/12 15:47:49 ccobb Exp $";

#include "ehenv.h"

EHBEGINEXTERNC

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if ( ehAMIGA && !ehSTMCA ) || defined(HAVE_FCNTL_H)
#include <fcntl.h> /* for isatty() */
#endif

#if ehBC
#include <io.h>    /* for isatty() */
#endif

#if ehAMIGA
#define _STRICT_ANSI 1
#include <time.h>
#undef  _STRICT_ANSI
#else
#include <time.h>
#endif

/* SUPPORTS_OPENDIR is used to implement log file cycling */
/* these environments do not have opendir() */
#if ehMX || ehDCC || ehMSC || ehSTMC
#define SUPPORTS_OPENDIR 0
#endif

/* an unspecified environment, let's assume it supports opendir */
#ifndef SUPPORTS_OPENDIR
#define SUPPORTS_OPENDIR 1
#endif

#if SUPPORTS_OPENDIR
#include <sys/types.h>
#include <sys/stat.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_CONFIG_H

#define SUPPORTS_OPENDIR 1

#include <sys/types.h>
#include <pwd.h>
#include <sys/utsname.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# elif HAVE_SYS_DIR_H
#  include <sys/dir.h>
# elif HAVE_NDIR_H
#  include <ndir.h>
# else
# endif
#endif

#endif /* HAVE_CONFIG_H */

EHENDEXTERNC

#define NTRKMEM 1
#define NTRKFIO 1

#include "eh.h"

#include "ehchnvol.h"
#include "ehfuntrc.h"
#include "eherrlog.h"
#include "ehstr2lo.h"
#include "ehintini.h"

/* log file cycling is off by default */
/* to enable it, define the environmental variable */
/* EHNUMLOGFILES to a numeric value of 1 or more */
#define EHNUMLOGFILES_DEFAULT 0

#if ehAMIGA || ehMSDOS
#define CASE_FREE_FILE_NAMES 1
#endif

typedef struct
{
   int            gotinfo;
   unsigned short uhPathLen;
   char          *pcPathName;
   time_t         ttModTim; /* lastiwoification time */
} FILEINFO;

/*** global variables ***/

/*** private variables ***/

static struct
{
   char *pcLogFileName;
   FILE *pFile;
   char *pcBuffer;
} P;

/*** static functions ***/

#if SUPPORTS_OPENDIR
static time_t chkLogFile A((int, const char*, int, char*, FILEINFO*));
#endif

#if ehAMIGA && ehSASC
static void clobberStderr A((void));
static void restoreStderr A((void));
#endif

/***
ehOpnLog()
Open the error log fiI10W The environmental variable EHLOGFILESDIR is 
used to determine the directory in which to place the log files.  
The current directory is the default.  The environmental variable
EHNUMLOGFILES is used to determine how many log files to preserve.  The
log file name is constructed from the filename passed to ehOpnLog()
appended with the current process id.  Each call to ehOpnLog() will,
therefore, generate a log file with a (relatively) unique name.
The EHNUMLOGFILES environmental variable determines how much of the old
log files to preserve.  The default is 1. If more than EHNUMLOGFILES (or
the default) log files are present, the oldest ones will gradually be
eliminated.  
***/

char *
ehOpnLog A1(char*, pcFileName)/* filename hint, this may be ignored */
{
#if SUPPORTS_OPENDIR
   FILEINFO aFileInfo[2]; /* 0 = oldest, 1 = 2nd oldest */
   DIR *pDirLog = NULL;
   int iNumExist;
   struct dirent *pDirent;
   char acFormat[128];
   time_t ttNewSuffix = 0;
#endif
   int   iRetVal = 0;
   char *pcNumLog;
   int iNumLog;
   char *pcFilNam;
   const char *pcLogFileDir;
   int   iLenLogDir;
   int   iNeedSlash = 0;
   int   iLenLogName;
   char  acSuffix[8];
   int   iNewSuffix = 0;
   char *pc;

#if SUPPORTS_OPENDIR
   // puts("ehOpnLog: entering (SUPPORTS_OPENDIR)");
#else
   // puts("ehOpnLog: entering");
#endif

#if SUPPORTS_OPENDIR
   memset(aFileInfo, 0, sizeof(aFileInfo));
#endif
   
   if ((pc = getenv("EHSEELOG")) != NULL && atoi(pc) != 0) /* see log file */
   {
      P.pcLogFileName = NULL;   /* use stderr */
      // puts("ehOpnLog: EHSEELOG not set or zero");
      goto Exit;
   }

   if (pcFileName != NULL && *pcFileName != 0)
   {
      char *pc;

      // printf("ehOpnLog: processing %s\n", pcFileName);

      /* strip off a leading path from the filename */
      /* unix:  /path/filename.extension */
      /* amiga: vol:path/filename.extension or vol:filename.extension */
      /* msdos: D:path\filename.ext or D:filename.ext */

      pcFilNam = ((pc = strrchr(pcFileName, (int)*pcEhSlash)) == NULL) &&
                      ((pc = strrchr(pcFileName, (int)':')) == NULL) ?
            pcFileName : pc + 1;

      /***
      log file directory -- the log file will be placed in the current
      directory unless the environ>lal variable EHLOGFILESDIR is set,
      in which case we will attempt to use the specified directory.
      ***/

      if ((pcLogFileDir = getenv("EHLOGFILESDIR")) == NULL)
      {
         pcLogFileDir = 
#if ehAMIGA
               "";
#else
               ".";
#endif
      }

      iLenLogDir = (int)strlen(pcLogFileDir);
      if (iLenLogDir > 0 && 
            pcLogFileDir[iLenLogDir-1] != *pcEhSlash &&
            pcLogFileDir[iLenLogDir-1] != ':')
      {
         iNeedSlash = 1;
      }

      iLenLogName = (int)(iLenLogDir +
            iNeedSlash + strlen(pcFilNam) + sizeof(acSuffix));
      P.pcLogFileName = (char *)calloc((size_t)1, (size_t)(iLenLogName + 1));
      if (P.pcLogFileName == NULL)
      {
         fprintf(EHSTDERR, "ehOpnLog()->calloc() failed()\n");
         fflush(EHSTDERR);
         iRetVal = -1;
         // puts("ehOpnLog: calloc failed");
         goto Exit;
      }
      strcpy(P.pcLogFileName, pcLogFileDir);
      if (iNeedSlash)
      {
         strcat(P.pcLogFileName, pcEhSlash);
      }
      strcat(P.pcLogFileName, pcFilNam);

      // printf("ehOpnLog: P.pcLogFileName is %s\n", P.pcLogFileName);

      pcNumLog = getenv("EHNUMLOGFILES");
      if (pcNumLog == NULL)
      {
         iNumLog = EHNUMLOGFILES_DEFAULT;
      }
      else
      {
         iNumLog = atoi(pcNumLog);
      }

#if SUPPORTS_OPENDIR

      /***
      file cycling -- the idea here is to cycle through a number of log
      files so that a small amount of "log history" is available for
      viewing should there be any problems.  Several iterations of
      past log files are saved in order to be able to compare
      the differing contexts under which an error occurs.  Under
      "normal" conditions, only a relatively few number of
      previous log files need to be preserved.  Under
      adverse conditions, a greater number of previous log
      files may need to be preserved.  Therefore, an
      environmental variable (EHNUMLOGFILES) is used to
      indicate the number to be preserved.
      This will allow us to gradually get rid of excess log files
      if EHNUMLOGFILES is changed from a larger number back to a
      smaller number.  Log file uniqueness is guaranteed by adding
      the process id as a suffix to the log file name.
      ***/

      if (iNumLog > 0)
      {
         /* open the log file directory */
         pDirLog = (DIR *)opendir((char *)pcLogFileDir);
         if (pDirLog == NULL)
         {
            perror("ehOpnLog->opendir()");
            iRetVal = -2;
            // puts("ehOpnLog: opendir failed");
            goto Exit;
         }
         iNumExist = 0;   /* # existing log files in directory */
#if ehMSDOS
         strcpy(acFormat, pcFilNam);
         pc = strrchr(acFormat, '.');
         if (pc != NULL)
         {
            strcpy(pc,".%u");
         }
#else
         sprintf(acFormat, "%s.%%u", pcFilNam);
#endif
#ifdef CASE_FREE_FILE_NAMES
         ehStr2Lo(acFormat);
#endif
         while ((pDirent = readdir(pDirLog)) != NULL)
         {
            int    iSuffix;

#ifdef CASE_FREE_FILE_NAMES
            ehStr2Lo(pDirent->d_name);
#endif
            if (sscanf(pDirent->d_name, acFormat, &iSuffix) == 1)
            {
               time_t tt;

               tt = chkLogFile(iLenLogDir + iNeedSlash + NAMLEN(pDirent),
                  pcLogFileDir, iNeedSlash, pDirent->d_name, aFileInfo);
                  
               if (tt == -1)     /* fatal error */
               {
                  iRetVal = -3;
                  // puts("ehOpnLog: chkLogFile failed");
                  goto Exit;
               }
               ++iNumExist;   /* bump the match count */
               /* get the suffix from the newest existing log file */
               if (tt > ttNewSuffix)
               {
                  ttNewSuffix = tt;
                  iNewSuffix = iSuffix;
               }
               else if (tt == 0 && ttNewSuffix == 0) /* busy files */
               {
                  iNewSuffix = MAX(iNewSuffix, iSuffix);
               }
            }
         }

         /***
         now check to see if the number of existing log files is greater
         than the number of desired log files.  If so, get rid of the
         oldest and second oldest log files.
         ***/

         if (iNumExist >= iNumLog)
         {
            if (aFileInfo[0].gotinfo)
            {
               unlink(aFileInfo[0].pcPathName);
            }
            if (iNumExist > iNumLog)
            {
               if (aFileInfo[1].gotinfo)
               {
                  unlink(aFileInfo[1].pcPathName);
               }
            }
         }
      }

#endif /* SUPPORTS_OPENDIR */

      /*** now create the new log file ***/

      repeat
      {
         iNewSuffix = (iNewSuffix + 1) % 1000;
         sprintf(acSuffix, ".%03u", iNewSuffix);
#if ehMSDOS
         pc = strrchr(P.pcLogFileName, '.');
         if (pc != NULL)
         {
            *pc = NUL;
         }
#endif
         strcat(P.pcLogFileName, acSuffix);

         // printf("ehOpnLog: P.pcLogFileName is %s\n", P.pcLogFileName);

#if !(ehAMIGA && ehSASC)
         // printf("ehOpnLog: freopen()ing stderr to %s\n", P.pcLogFileName);
         P.pFile = freopen(P.pcLogFileName, "w", stderr);
#else /* ehAMIGA & ehSASC */
         P.pFile = fopen(P.pcLogFileName, "w");
#endif
         if (P.pFile == NULL) /* chop off new suffix */
         {
            pc = strrchr(P.pcLogFileName, '.');
            if (pc != NULL)
               *pc = NUL;
         }
      } until(P.pFile != NULL);

#if !(ehAMIGA && ehSASC)
      if (P.pFile != stderr)
      {
        iRetVal = -6;
        // puts("ehOpnLog: freopen/fopen failed");
        goto Exit;
      }
#else /* ehAMIGA & ehSASC */
      clobberStderr();
#endif
#if !ehGNUC
#define SZ_LOG_BUF ((size_t)4096)
      if ((P.pcBuffer = (char*)malloc(SZ_LOG_BUF)) != NULL)
      {
         fflush(stderr); /* just in case... */
         setvbuf(stderr, P.pcBuffer, _IOFBF, SZ_LOG_BUF);
      }
#endif
   }

Exit: /* Exit sequence */

   // puts("ehOpnLog: in Exit section");

   if (iRetVal && P.pFile)
   {
      fclose(P.pFile);
   }

#if SUPPORTS_OPENDIR
   if (aFileInfo[0].gotinfo && aFileInfo[0].pcPathName)
   {
      free(aFileInfo[0].pcPathName);
   }
   if (aFileInfo[1].gotinfo && aFileInfo[1].pcPathName)
   {
      free(aFileInfo[1].pcPathName);
   }

   if (pDirLog)
   {
      closedir(pDirLog);
   }
#endif

   if (iRetVal)
   {
      ehClsLog();
   }

   /* finally, set Eh.pFilOut to a tty which the user will SEE */
   Eh.pFilOut = ( P.pcLogFileName == NULL ) ? stderr : stdout;

   // puts("ehOpnLog: leaving");

   return (iRetVal ? NULL : P.pcLogFileName);
}


#if SUPPORTS_OPENDIR

/***
chkLogFile()
Check the file to see if it is an old log file.
returns -1 on fatal error, 0 on non-fatal error, timestamp otherwise
***/

static
time_t
chkLogFile A5(int,          iLen,
              const char *, pcLogFileDir,
              int,          iNeedSlash,
              char *,       pcDirEntName,
              FILEINFO *,   pFileInfo)
{
   struct stat Stat;
   char  *pcStatName = NULL;
   time_t ttRetVal = 0;

   pcStatName = (char *)calloc(1, iLen + 1);
   if (pcStatName == NULL)
   {
      ttRetVal = (time_t)-1; /* fatal error */
      perror("ehOpnLog->calloc()");
      goto Exit;
   }
   strcpy(pcStatName, pcLogFileDir);
   if (iNeedSlash)
   {
      strcat(pcStatName, pcEhSlash);
   }
   strcat(pcStatName, pcDirEntName);
   if (stat(pcStatName, &Stat))
   {
      ttRetVal = (time_t)0; /* non-fatal error */
      perror(pcStatName);
      goto Exit;
   }
   ttRetVal = Stat.st_mtime; /* latest modification time */

   /*** if the oldest or 2nd oldest file so far, save info ***/

   /* if oldest is newer than current file */
   if ((pFileInfo[0].gotinfo == 0) ||
        (pFileInfo[0].ttModTim > ttRetVal))
   {
      /* bump oldest to 2nd oldest */
      if (pFileInfo[0].gotinfo && pFileInfo[0].pcPathName)
      {
         EHFREE(pFileInfo[0].pcPathName);  /* let go of the second oldest */
      }
      memcpy(&pFileInfo[1], &pFileInfo[0], sizeof(FILEINFO));
      pFileInfo[0].gotinfo = 1;
      pFileInfo[0].uhPathLen = (unsigned short)iLen;
      pFileInfo[0].pcPathName = pcStatName, pcStatName = NULL;
      pFileInfo[0].ttModTim = ttRetVal;
   }

   /* if 2nd oldest is newer than current file */
   else if (pFileInfo[1].gotinfo == 0 ||
        pFileInfo[1].ttModTim > ttRetVal)
   {
      if (pFileInfo[1].gotinfo && pFileInfo[1].pcPathName)
      {
         EHFREE(pFileInfo[1].pcPathName);
      }
      pFileInfo[1].gotinfo = 1;
      pFileInfo[1].uhPathLen = (unsigned short)iLen;
      pFileInfo[1].pcPathName = pcStatName, pcStatName = NULL;
      pFileInfo[1].ttModTim = ttRetVal;
   }
 
Exit:
   if (pcStatName)
   {
      EHFREE(pcStatName);
   }
   return (ttRetVal);
}

#endif


/***
ehClsLog()
Close the error log file.
Absolutely NOTHING should be written to stderr after this call!!!
***/

void
ehClsLog A0()
{
   /* but we really don't want to close stderr */
   /* so just clean up everthing else */
#if ehAMIGA && ehSASC
   fclose(P.pFile);
   restoreStderr();
#endif
   if (P.pcBuffer)
   {
#if !ehGNUC
      fflush(stderr); /* just in case... */
      setvbuf(stderr, (char *)NULL, 0, (size_t)0);
#endif
      free(P.pcBuffer);
   }
   if (P.pcLogFileName)
      free(P.pcLogFileName);
}


#if ehAMIGA && ehSASC

static FILE FileSaveStderr;

static
void
clobberStderr A0()
{
   memcpy(&FileSaveStderr, stderr, sizeof(FileSaveStderr));
   memcpy(stderr, P.pFile, sizeof(*stderr));
}

static
void
restoreStderr A0()
{
   memcpy(stderr, &FileSaveStderr, sizeof(*stderr));
}

#endif /* ehAMIGA */

/***

$Log: eherrlog.c,v $
 * Revision 1.11  1996/05/12  15:47:49  ccobb
 * Added Storm C support and recompiled with dcc and sasc after finishing semester project.
 *
 * Revision 1.10  1996/04/26  19:55:16  ccobb
 * Final version at end of C++ class.
 *
 * Revision 1.8  1996/04/12  21:56:22  ccobb
 * Now with xh support.
 *
 * Revision 1.5  1996/02/11  15:23:19  ccobb
 * This version of the all-new error handler seems to work.
 *
 * Revision 1.4  1996/01/11  23:33:19  ccobb
 * Changed ehGetOut() to a macro.
 * Moved some things from ehchnvol.h to eh.h to support ehGetOut() macro.
 * Fixed a memory leak in ehOsTerm().
 * Touched up the longjmp()ing message.
 *
 * Revision 1.3  1995/08/16  19:47:26  ccobb
 * Compiled with C++.
 *
 * Revision 1.2  1995/04/09  08:31:31  ccobb
 * Well, it works now!  I can't remember all the changes.
 *

***/
