/***

ehfuntrc.c -- Error Handler, function tracing support

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char *rcsid =
       "$Id: ehfuntrc.c,v 1.18 1996/05/16 20:54:44 ccobb Exp $";

#include "ehenv.h"

EHBEGINEXTERNC
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
EHENDEXTERNC

#include "ehchnvol.h"
#include "ehfuntrc.h"
#include "ehintres.h"
#include "ehsigint.h"
#include "ehchkovr.h"
#include "ehercode.h"

EH      Eh = { "unknown" };    /* global variables */
EHSTATE EhSt;                  /* global state variables */

char    cEhPgmTrc = 0; /* internal program  tracing */
char    cEhFunTrc = 0; /* internal function tracing */
char    cDmpEhSt  = 0; /* dump the EhSt list */

#define INTERRUPTED()\
 (EhSt.yCtrlC && EhSt.yChain && !EhSt.yInhibit && !EhSt.yShield)
#define WANTTOJUMP()  (EhSt.yChain)
#define HAVETOJUMP()  (EhSt.yChain && !EhSt.yInhibit && !EhSt.yShield)

/*** local functions ***/

static void dmpEhSt    A((EHSTATE *pEhSt));

#ifndef ehEXCEPTIONHANDLING
static void longJump   A((const char *pcFunc));
static void restoreJump A((EHSTATE*pEhSvPar, EHSTATE*pEhSvCat));
#endif /* !ehEXCEPTIONHANDLING */


/***
ehIndent()
Space over according to the current nesting level.
***/

#ifndef ehIndent /* see eh.h */
void 
ehIndent A1(FILE *, pFile)
{
   fprintf(pFile, "%*.*s", EhSt.hLev * 3, EhSt.hLev * 3, "");
   fflush(pFile);
} /* ehIndent() */
#endif /* ehIndent */


/***
ehLogRet()
Log the function-return.
***/

void
ehLogRet A4(int,          iChan,
            const char *, pcFile,
            const char *, pcFunc,
            const int,    iLineExit) /* {0|exit line} */
{
   EHSTATE *pEhSvPar = EhSt.pEhSv;        /* for local use */
   int iInterrupted;
   int iGetFun;
   EHSTATE *pEhSvCat = pEhSvPar;          /* point to parent */

   --EhSt.hLev; /* for indentation purposes */

   /* find the EHSTATE of the most recent catcher */
   while (pEhSvCat->pEhSv && pEhSvCat->yCatch == 0)  /* most recent catcher */
      pEhSvCat = pEhSvCat->pEhSv;

   /*** now do some logging ***/

   iInterrupted = 
         (EhSt.yCtrlC && EhSt.yChain && !EhSt.yShield && !EhSt.yInhibit);
   if ((iGetFun = ehGetFun(iChan)) != 0)
   {
      ehIndent(EHSTDERR);  
      fprintf(EHSTDERR, "} /""* %s %s%s",
            pcFile,
            pcFunc?pcFunc:"",
            pcFunc?"() ":"");
      if (iLineExit)
         fprintf(EHSTDERR, "%d ", iLineExit);
      fprintf(EHSTDERR, "*""/");
   }
   else if (EhSt.scError || iInterrupted)
   {
      ehIndent(EHSTDERR);  
      fprintf(EHSTDERR, "%s %s%s",
            pcFile,
            pcFunc?pcFunc:"",
            pcFunc?"() ":"");
      if (iLineExit)
         fprintf(EHSTDERR, "%d ", iLineExit);
   }

   /* Ctrl-C */
   if (iInterrupted) /* same as in ehFunEnt() */
   {
      fputs(" -- interrupted", EHSTDERR);
   }
   else if (EhSt.scError)
   {
      fputs(" -- failed", EHSTDERR);
   }

   if (iGetFun || iInterrupted || EhSt.scError)
   {
      fputs("\n", EHSTDERR);
   }
   fflush(EHSTDERR);

} /* ehLogRet() */


/***
ehChkCtrlC()
Check to see if an interrupt signal has
occured and if so, propogate it into the
Error Handler state variable.

Can't set scError here otherwise we would get
"failed" messages even when a CtrlC wasn't
fully processed and function really was successful.
***/

void 
ehChkCtrlC A0()
{


   /***
      We had to add EhSt.yInhibit to the Control-C check otherwise
      the Error Handler would be left in an inconsistent state.
      Namely, if Ctrl-C chaining were activated while inhibition
      was set, it would remain suppressed until inhibition is
      lifted which would normally be in the ehFunRet().  This
      could happen in an unexpected (and more importantly, 
      unanticipatable) way which would be particularly nasty
      if returning a resource handle.  Because chaining started
      suddenly, a normal return would not be possible and the
      resource handle would not be returned.  In fact it would be
      lost altogether.  This can be prevented by "detecting" a
      Control C only when it can be fully handled.
   ***/
   if (Eh.yIntSig && !Eh.yDisable && !EhSt.yInhibit && !EhSt.yShield)
   {
      if (Eh.Er.hCode >= 0) Eh.Er = ehUnHash(EH_CTRLC_HASH);
      fputs("\n^C", EHSTDOUT);
      fflush(EHSTDOUT);
      if (EHSTDOUT != EHSTDERR)
      {
         fputs("\n^C", EHSTDERR);
         fflush(EHSTDERR);
      }
      EhSt.yCtrlC = 1;
      EhSt.yChain = 1;
      Eh.yIntSig  = 0;
   }
} /* ehChkCtrlC */


/***
dmpEhSt()
This function dumps the EhSt structure
and the list of linked structures.
***/

static
void
dmpEhSt A1(EHSTATE *, pEhSt)
{
   fprintf(EHSTDERR, "\nEhSt %06.6p ", pEhSt);
   fprintf(EHSTDERR, "pEhSv %06.6p ",  pEhSt->pEhSv);
   fprintf(EHSTDERR, "%10.10s ",       pEhSt->pcFile ? pEhSt->pcFile:"<file>");
   fprintf(EHSTDERR, "%-10.10s " ,     pEhSt->pcFunc ? pEhSt->pcFunc:"<func>");
   fprintf(EHSTDERR, "Lev % 2d ",      pEhSt->hLev);
   fprintf(EHSTDERR, "\n\t");
   fprintf(EHSTDERR, "Catch %d ",      pEhSt->yCatch);
   fprintf(EHSTDERR, "Chain %d ",      pEhSt->yChain);
   fprintf(EHSTDERR, "CtrlC %d ",      pEhSt->yCtrlC);
   fprintf(EHSTDERR, "Inhib %d ",      pEhSt->yInhibit);
   fprintf(EHSTDERR, "Shld %d ",       pEhSt->yShield);
   fprintf(EHSTDERR, "scErr % 2d ",    pEhSt->scError);
   if (pEhSt->pEhSv)
      dmpEhSt(pEhSt->pEhSv);
   else
      fprintf(EHSTDERR,"\n");
   fflush(EHSTDERR);
} /* dmpEhSt */


/***
ehCallStack()
This function dumps the current call stack.
***/

void
ehCallStack A1(
   FILE *, pFile) /* output file */
{
   EHSTATE *pEhSt = &EhSt;
   ehIndent(pFile);
   fputs("---------- call stack ---------------------\n",
         pFile);
   do
   {
      if (pEhSt->pEhSv)
      {
         ehIndent(pFile);
         fprintf(pFile, "%15.15s %s%s%d (%s)\n",
            pEhSt->pcFile,
            pEhSt->pcFunc?pEhSt->pcFunc:"",
            pEhSt->pcFunc?"() ":"",
            pEhSt->iLine,
            pEhSt->yCatch ? "catcher" : "tracer");
      }
   } while ((pEhSt = pEhSt->pEhSv) != 0);
   ehIndent(pFile);
   fputs("-------------------------------------------\n",
         pFile);
} /* ehCallStack */


/*** Support for Standard C error handling only ***/


#ifndef ehEXCEPTIONHANDLING


/***
ehPostJmp()
This function gets called immediately after the longjmp()
in the context of the catching function.
***/

void 
ehPostJmp A1(const char *, pcFunc)
{
   fprintf(EHSTDERR, "to Cleanup Section ");
   if (pcFunc) fprintf(EHSTDERR, "in %s() ", pcFunc);
   fprintf(EHSTDERR, "***");
   if (cDmpEhSt) dmpEhSt(&EhSt);
   fprintf(EHSTDERR, "\n");
   fflush(EHSTDERR);
} /* ehPostJmp */


/***
ehFunEnt()
Function-entry tracing support.
Chaining support has been moved to ehCatch().
***/

void
ehFunEnt A6(int,          iChan,
            const char *, pcFile,
            const char *, pcFunc,
            const int,    iLine,
            int,          iChkOvr,
            EHSTATE *,    pEhSv) /* dummy to silence compiler */
{
   int iInterrupted;
   int iGetFun;
   bool caller_is_catcher  = (bool)EhSt.yCatch;
   bool just_entering_lev0 = (bool)(EhSt.hLev == 0);

   // puts("ehFunEnt: entering");

   /*** if requested, check for memory overwrites ***/

   if (iChkOvr) ehChkOvr(pcFile, pcFunc, iLine, &ehIntRes);

   /*** CTRL-C longjmp() ***/

   ehChkCtrlC(); /* Eh.iIntSig --> EhSt.{Er.lNo,CtrlC,Chain} */

   /*** now do some logging ***/

   if ((iGetFun = ehGetFun(iChan)) != 0)
   {
      ehIndent(EHSTDERR);

#ifdef EHSIMCTRLC /* compiler-line request for CTRL-C simulation */
      if (Eh.ulCtrlCVal + 1)
         fprintf(EHSTDERR, "%ld ", Eh.ulCtrlCVal + 1);
#endif

      fprintf(EHSTDERR, "%s %s%s%d {", /* "}" */
            pcFile,
            pcFunc?pcFunc:"",
            pcFunc?"() ":"",
            iLine);
      fflush(EHSTDERR);
   }

   /***
      NOTE that the CTRL-C state that we just processed,
      is now considered to be part of the parent's
      environment which, if we longjmp() will become
      part of the catcher's environmemt.
   ***/

   /*** finish logging activity, taking CTRL-C into account ***/

   if ((iInterrupted = INTERRUPTED()) != 0)
   {
      if (iGetFun)
      {
         fputs("   -- interrupted", EHSTDERR);
         if (!just_entering_lev0)
         {
            fputs("\n", EHSTDERR);
            ehIndent(EHSTDERR);                     /* balance out the "{" */
            fprintf(EHSTDERR, "} /""* %s %s%s%d *""/",
                  pcFile,
                  pcFunc?pcFunc:"",
                  pcFunc?"() ":"",
                  iLine);
         }
      }
      else /* no function tracing: display interrupted function & call stack */
      {
         EHSTATE *pEhSt = &EhSt;
         ehIndent(EHSTDERR);
         fprintf(EHSTDERR, "%s %s%s%d -- interrupted\n", 
               pcFile,
               pcFunc?pcFunc:"",
               pcFunc?"() ":"",
               iLine);
         ehCallStack();
      }
      fflush(EHSTDERR);
   }
   if (iGetFun)
   {
      fputs("\n", EHSTDERR);
      fflush(EHSTDERR);
   }

   /*
      This part used to be in ehCatch().
      We moved it here to accomodate ehdivsrc
      which doesn't call PGM_ENTER().
   */ 

   if (just_entering_lev0) /* set up dummy EhSt.pEhSv */
   {
      pEhSv->pcFile   =
      pEhSv->pcFunc   = "theAbyss";
      pEhSv->iLine    = -1;
      pEhSv->pEhSv    = NULL; /* bottom of stack */
      pEhSv->hLev     = 0;
      pEhSv->yCatch   = 0; /* shouldn't this be here? */
      pEhSv->yInhibit = 0;
      pEhSv->yShield  = 1; /* never jump out of main */
      pEhSv->scError  = 0;
      pEhSv->yCtrlC   = 0;
      pEhSv->yChain   = 0;

      EhSt.pEhSv = pEhSv;     /* need to do this *before* longjmp() */
      memcpy(pEhSv->JmpBuf, EhSt.JmpBuf, sizeof(pEhSv->JmpBuf));
   }

   /***
      if we have been interrupted, we will try not to execute
      the function we are enterring.  Unless restrictive modes
      are set (yInhibit, yShield), we will longjmp() back to the
      most recent catcher function.  FUN_ENTER() therefore behaves
      as a checkpoint for interrupt (CTRL-C) handling.  If restrictive
      modes are set (such as yInhibit or yShield) we will not longjmp()
      and we will defer CTRL-C handling to FUN_RETURN()/FUN_LEAVE()
      (ehFunRet()) or to the next call to ehFunEnt()
   ***/

   if (iInterrupted) /* CTRL-C + HAVETOJUMP() */
   {
      if (cDmpEhSt) dmpEhSt(&EhSt);

      EhSt.scError = -1; /* resolve the CTRL-C to a real error */

      if (!caller_is_catcher)
      {
         EHSTATE *pEhSvCat = EhSt.pEhSv;     /* point to parent */
         while (pEhSvCat->pEhSv && pEhSvCat->yCatch == 0)
            pEhSvCat = pEhSvCat->pEhSv;      /* find catcher */

         EhSt.pcFile = pEhSvCat->pcFile;
         EhSt.pcFunc = pEhSvCat->pcFunc;
         EhSt.iLine  = pEhSvCat->iLine;
         EhSt.yCatch = 1;                    /* we are jumping to a catcher */
         EhSt.hLev   = pEhSvCat->hLev;  /* catcher's nesting level */
         EhSt.pEhSv  = pEhSvCat->pEhSv; /* catcher's EHSTATE variable address */
      }
      else if (just_entering_lev0)
      {
         EhSt.pcFile = pcFile;
         EhSt.pcFunc = pcFunc;
         EhSt.iLine  = iLine;
         ++EhSt.hLev;        /* entering next nesting level */
      }

      longJump(pcFunc); /*** attempt to abort this function and start chaining ***/
   }

   /*** we were not interrupted or could not chain: proceed as normal ***/

   /*** Save the current Error Handler state. ***/

   if (!just_entering_lev0) /* save Error Handler state */
   {
      /* save values of our parent */
#if 1
      memcpy(pEhSv, &EhSt, sizeof(*pEhSv));
#else
      pEhSv->pcFile   = EhSt.pcFile;
      pEhSv->pcFunc   = EhSt.pcFunc;
      pEhSv->iLine    = EhSt.iLine;
      pEhSv->yCatch   = EhSt.yCatch;
      pEhSv->yChain   = EhSt.yChain;
      pEhSv->yCtrlC   = EhSt.yCtrlC;
      pEhSv->yInhibit = EhSt.yInhibit;
      pEhSv->yShield  = EhSt.yShield;
      pEhSv->scError  = EhSt.scError;
      pEhSv->pEhSv    = EhSt.pEhSv;
#endif
   }

   /***
      Update the Error Handler state.

      If we are being called from main(), this function
      gets called *after* ehCatch(), otherwise it gets
      called *before* ehCatch().  This is determined
      by examining hLev.
   ***/

   /*** establish new state ***/

   EhSt.pEhSv = pEhSv; /* save address of local EHSTATE variable */

   if (!just_entering_lev0)
   {
      EhSt.yCatch = 0;
   }

   EhSt.pcFile = pcFile;
   EhSt.pcFunc = pcFunc;
   EhSt.iLine  = iLine;
   EhSt.yChain = 0;    /* clear the chaining request bit */
   EhSt.yCtrlC = 0;    /* no interrupts have (yet) occurred in this function */
   ++EhSt.hLev;        /* entering next nesting level */
   EhSt.scError = 0;    /* no errors have (yet) occurred in this function */

   if (cDmpEhSt) dmpEhSt(&EhSt);

   // puts("ehFunEnt: leaving");

} /* ehFunEnt() */


/***
ehCatch()
This function enables catching.  It is called via the FUN_CATCH()
macro.  If FUN_CATCH() is not called, then a function is not
capable of receiving a thrown exception.  This has the same effect
as being shielded.  In other words, if FUN_CATCH() is not called,
and a child function wants to throw an exception, it won't be able
to.  This has the same effect as whenever we cannot throw an exception:
further chaining is stopped.
***/

int
ehCatch A1(EHSTATE *,    pEhSv) /* save the previous state */
{
   bool just_entering_lev0 = (bool)(EhSt.hLev == 0); /* not yet incremented by ehFunEnt() */

   // puts("ehCatch: entering");

   /*** we are a catcher ***/

   EhSt.yCatch  = 1; /* we are a catcher */
   EhSt.yShield = 0; /* let us catch! */

   /***
      Update the Error Handler state.

      If we are being called from main(), this function
      gets called *before* ehFunEnt(), otherwise it gets
      called *after* ehFunEnt().  This is determined
      by examining hLev.
   ***/

   if (!just_entering_lev0) /* setjmp() has been called at *least* once */
      memcpy(pEhSv->JmpBuf, EhSt.JmpBuf, sizeof(pEhSv->JmpBuf));

   // puts("ehCatch: leaving");

   return(0);
} /* ehCatch() */


/***
ehFatalJmp()
This function is called from
   ehPError() - in EHTHROW()
This function longjmp()s to the nearest ExitSequence.
This ExitSequence may be in the current function
(if the current function is a catcher) or somewhere
up the call stack.  If the current function is a
catcher, we can ignore inhibition and shielding and
simply jump to the current ExitSequence.  If the current
function is not a catcher, we may be inhibited or
shielded by a functino further up the call stack.
***/

void
ehFatalJmp A0()
{
   EHSTATE *pEhSv = &EhSt;         /* point to current function */
   EHSTATE *pEhSvPar = EhSt.pEhSv; /* point to parent */
   EHSTATE *pEhSvCat = pEhSv;      /* point to most recent catcher */
   byte yShield;     /* most restricting condition */

   if (cDmpEhSt) dmpEhSt(&EhSt);

   /* find the EHSTATE of the most recent catcher */
   while (pEhSvCat->pEhSv && pEhSvCat->yCatch == 0)  /* find nearest catcher */
      pEhSvCat = pEhSvCat->pEhSv;

   /*** can we just goto a local ExitSequence? ***/

   if (pEhSv == pEhSvCat) /* current function is a catcher */
   {
#if 0
      longJump(pcFunc); /* makes debugging easier: stops in ehPostJmp() */
#endif
      return;           /* do local goto ExitSequence */
   }
 
   /*** if allowed, jump to catcher ***/

   /* see if we are allowed to longjmp() */
   yShield  = (byte)(pEhSv->yShield || pEhSvPar->yShield || pEhSvCat->yShield); 
   if (EhSt.yChain && !EhSt.yInhibit && !yShield)
   {
      restoreJump(pEhSvPar, pEhSvCat);   /* restore catcher's values and longjmp() */
   }
} /* ehFatalJmp() */


/***
ehExitJmp()
This function is called if user calls exit().
First ehAtExit() is called which determines
whether chaining is appropriate.  If so, it
calls this function to attempt to initiate
chaining.  If we are not allowed to chain,
we can just return, at which point ehAtExit()
will do system cleanup.  Otherwise, we can
commense chaining and will eventually end up
back in main() where final system cleanup
will be performed.
***/

void
ehExitJmp A0()
{
   EHSTATE *pEhSv = &EhSt;         /* point to current */
   EHSTATE *pEhSvPar = EhSt.pEhSv; /* point to parent */
   EHSTATE *pEhSvCat = pEhSv;      /* point to catcher */

   if (cDmpEhSt) dmpEhSt(&EhSt);

   /* find the EHSTATE of the most recent catcher */
   while (pEhSvCat->pEhSv && pEhSvCat->yCatch == 0)  /* find most recent catcher */
      pEhSvCat = pEhSvCat->pEhSv;

   /*** may we start chaining ? ***/

   /* this is a last-ditch attempt */
   if (/*EhSt.yChain &&*/ !EhSt.yInhibit && !pEhSvCat->yShield)
   {
      restoreJump(pEhSvPar, pEhSvCat);   /* restore values and longjmp() */
   }
} /* ehExitJmp() */


/***
ehFunRet()
Function-return tracing support.
***/

void
ehFunRet A5(int,          iChan,
            const char *, pcFile,
            const char *, pcFunc,
            const int,    iLine,
            int,          iChkOvr)
{
   EHSTATE *pEhSvPar = EhSt.pEhSv;        /* for local use */
   bool caller_is_catcher = (bool)EhSt.yCatch;  /* for local use */
   EHSTATE *pEhSvCat = pEhSvPar;          /* point to parent */

   /*** if requested, check for memory overwrites ***/

   if (iChkOvr) ehChkOvr(pcFile, pcFunc, iLine, &ehIntRes);

   fflush(EHSTDERR);

   /*** now do some logging ***/

   ehLogRet(iChan, pcFile, pcFunc, iLine);

   /***
      At this point, we need to examine the current state
      of the error handler and determine whether we are going to:
         o return normally, or
         o longjmp() back to previous catching function.
   ***/ 

   if (cDmpEhSt) dmpEhSt(&EhSt);

   /* find the EHSTATE of the most recent catcher */
   while (pEhSvCat->pEhSv && pEhSvCat->yCatch == 0)  /* find most recent catcher */
      pEhSvCat = pEhSvCat->pEhSv;

   /***
      yInhibition --
      Much of the time when we call ehFunRet(), we will be in the ExitSequence
      of a function -- either through normal execution or because of an exception.
      In this case, yInhibition will always be set.  What we want is whether our
      parent is inhibiting us.  Therefore, we get the parent's value before
      testing HAVETOJUMP().

      This is complicated by the occurance of a Ctrl-C during the ExitSequence.
      Because of the inhibition, it will not be processed immediately.  It will
      remain pending until reaching ehFunRet().  If the parent does not have
      inhibition set, we would be tempted to start the chain now.  However, this
      gives the current function no chance to respond to the Ctrl-C.  A resource
      could have been obtained in the current function whose handle is being passed
      back to the parent.  If we process the Ctrl-C here, a normal return does not
      take place, and therefore the parent never sees the handle.  However, the
      current function never learns about the Ctrl-C until FUN_RETURN() -- where it
      is to late to recover. 
      
      The parent's inhibition state should be restored *after* the longjmp()
      test is made to ensure that a normal return takes place -- giving resource
      release responsibility to the parent.

      WE MUST set it before checking whether to longjmp() because otherwise we can
      never longjmp from a catcher function: the local ExitSequence will have set
      inhibition.

      WE MUST NOT set it before checking whether to longjmp() because then we can
      loose resource responsibility: suppressed chaining (e.g., from a control C)
      will now become activated and returning a successfully obtained resource
      handle will be impossible.  The resource handle will be lost.
   ***/
   EhSt.yInhibit = pEhSvPar->yInhibit; /* possibly turn off inhibition */
   EhSt.yShield  = (byte)(pEhSvPar->yShield || pEhSvCat->yShield); /* most restricting condition */
   if (HAVETOJUMP())
   {
      if (EhSt.yCatch)
         memcpy(EhSt.JmpBuf, pEhSvCat->JmpBuf, sizeof(EhSt.JmpBuf));

      restoreJump(pEhSvPar, pEhSvCat); /* restore values and longjmp() */
   }

   /*** normal function return ***/

   if (caller_is_catcher)
      memcpy(EhSt.JmpBuf, pEhSvCat->JmpBuf, sizeof(EhSt.JmpBuf));

   /* restore parent's values */
   EhSt.pcFile   = pEhSvPar->pcFile;
   EhSt.pcFunc   = pEhSvPar->pcFunc;
   EhSt.iLine    = pEhSvPar->iLine;
   EhSt.yShield  = pEhSvPar->yShield;  /* possibly turn off shielding */
   EhSt.yCatch   = pEhSvPar->yCatch;
   EhSt.yChain  |= pEhSvPar->yChain;
   EhSt.yCtrlC  |= pEhSvPar->yCtrlC;
   EhSt.hLev     = pEhSvPar->hLev;     /* restore parent's nesting level */
   EhSt.scError |= pEhSvPar->scError;
   EhSt.pEhSv    = pEhSvPar->pEhSv;
} /* ehFunRet() */


/***
restoreJump()
This function assumes that it is OK to longjmp().
No checks are made.  It sets several state
variable according to the parent's and/or
catcher's values.  Then performs the longjmp().
??? Should shielding be reset here ???
Shielding does not prevent jumping from a
shielded function, only jumping to a 
shielded function.  Therefore, we may be
in a shielded function jumping to an un-
shielded function, in which case we should
reset yShield according to our destination.
***/

static
void
restoreJump A2(EHSTATE*,pEhSvPar, /* Parent */
              EHSTATE*,pEhSvCat) /* Catcher */
{
   const char *pcFunc = EhSt.pcFunc;

   if (EhSt.yCtrlC)
      EhSt.scError = -1; /* resolve the Control-C to an error */

   EhSt.pcFile   = pEhSvCat->pcFile;
   EhSt.pcFunc   = pEhSvCat->pcFunc;
   EhSt.iLine    = pEhSvCat->iLine;
   EhSt.yShield  = pEhSvCat->yShield; /* ??? */
   EhSt.yCatch   = 1;  /* we are jumping into a catcher */
   EhSt.yCtrlC  |= pEhSvPar->yCtrlC  || pEhSvCat->yCtrlC;  /* pick worst condition */
   EhSt.hLev     = pEhSvCat->hLev;                         /* restore catcher's nesting level */
   EhSt.scError |= pEhSvPar->scError || pEhSvCat->scError; /* pick most restricting condition */
   EhSt.pEhSv    = pEhSvCat->pEhSv;                        /* restore catcher's EHSTATE variable */

   longJump(pcFunc);
} /* restoreJump() */


/***
longJump()
Logs and performs a longjmp()...no checking.
***/

static
void
longJump A1(const char *, pcFunc) /* may be NULL */
{
   ehIndent(EHSTDERR);
   fputs("*** longjmp()ing", EHSTDERR);
   if (pcFunc) fprintf(EHSTDERR, " from %s()", pcFunc);
   fputs(" *** ", EHSTDERR);
   fflush(EHSTDERR);
#if 0 /* emit function exit messages for all skipped functions */
      {
         EHSTATE *pEhSt = &EhSt;
         fprintf(EHSTDERR, "\n");
         do
         {
            if (pEhSt->pEhSv)
            {
               ehIndent(EHSTDERR);
               fprintf(EHSTDERR, "->%s %s%s%d",
                  pEhSt->pcFile,
                  pEhSt->pcFunc?pEhSt->pcFunc:"",
                  pEhSt->pcFunc?"() ":"",
                  pEhSt->iLine);
               if (pEhSt->yCatch)
                  fputs(" (catcher)", EHSTDERR);
               fputs("\n", EHSTDERR);
            }
         } while ((pEhSt = pEhSt->pEhSv) != 0);
      }
#endif
   longjmp(EhSt.JmpBuf, (int)Eh.Er.lNo); /*** *** longjmp() *** ***/
} /* longJump() */


#endif /* !ehEXCEPTIONHANDLING */


/*** Support for true C++ Exception Handling ***/


#ifdef ehEXCEPTIONHANDLING


#include <time.h>

#include "ehcpuchk.h"
#include "eherrlog.h"
#include "ehvolume.h"
#include "ehavltre.h"
//#include "ehpgmtrc.h"

static char   cFinished;
static char  *pcLogFileName;
static char   acExecute[128];
static time_t ttStart;

static void xhThrow(const char *pcFunc);
static void restoreContext(EHSTATE*pEhSvPar);
static void xhAtExit();
static void xhExitThrow();

// EHFUNTRACE Default Contructor
EHFUNTRACE::EHFUNTRACE()
{
   // don't do nuttin'
} // EHFUNTRACE::EHFUNTRACE

// EHFUNTRACE Contructor: function tracer
EHFUNTRACE::EHFUNTRACE(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   int iChan,
   bool chkOvrwrt)
{
   ft_Init(pcFile, pcFunc, iLine, iChan, chkOvrwrt);
} // EHFUNTRACE::EHFUNTRACE

/***
Examine command-line arguments.
Extract EHVOLUME settings.
***/
void
EHFUNTRACE::ft_Args(
   int   *piArgc,
   char **ppcArgv)
{
   volatile int iArgc = *piArgc;

   for (++ppcArgv; --iArgc > 0; ++ppcArgv) /* skip over argv[0], then loop */
   {
      char *pc = *ppcArgv;   /* get next argument from command line */

      /* process Error Handler flags */
      if (memcmp(pc, "EHVOLUME=", 9) == 0)
      {
         ehVolume(pc + 9);

         /* take out EHVOLUME entry */
         --(*piArgc);
         while (iArgc-- > 0)    /* skip over argv[0], loop thru args */
         {
            *ppcArgv = *(ppcArgv +1); /* get next argument from command line */
            ++ppcArgv;
         }
         break;
      }
   }
} // EHFUNTRACE::ft_Args

/***
EHFUNTRACE initialization routine
This function contains the meat of the logic
that would normally be contained in the 
constructor.  The reason it is not in the
constructor is the EHFUNTRACE object must
be instantiated *before* the try block
begins in main() so that it is visible
throughout the whole scope.  This means
that a default constructor is necessary.
But it must be initilized *within* the
try block so that any errors that may
occur during initilization are caught
by the catch statement.
***/
void
EHFUNTRACE::ft_Init(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   int iChan,
   bool chkOvrwrt,
   int *piArgc,
   char **ppcArgv)
{
   ft_pcFile = pcFile;
   ft_pcFunc = pcFunc;
   ft_iLine = iLine;
   ft_iLineExit = 0;
   ft_iChan = iChan;
   ft_chkOvrwrt = chkOvrwrt;
   int iInterrupted;
   int iGetFun;
   bool caller_is_catcher  = (bool)(EhSt.yCatch != 0);
   bool just_entering_lev0 = (bool)(EhSt.hLev == 0);

   if (cEhFunTrc)
      printf("EHFUNTRACE::ft_Init: entering (entering %s, hLev %d)\n",
            pcFunc?pcFunc:"<func>", EhSt.hLev);

   /*** if first time through, initialize ***/

   if (Eh.yIniState == 0) // let's initialize, shall we?
   {
      ++Eh.yIniState;             // mainly to prevent recursion
      ft_PgmEnt(piArgc, ppcArgv); // program initialization (recursive!)
   }

   /*** check for memory overwrites ***/

   //puts("EHFUNTRACE::EHFUNTRACE: about to call ehChkOvr");
   if (chkOvrwrt) ehChkOvr(pcFile, pcFunc, iLine, &ehIntRes);
   //puts("EHFUNTRACE::EHFUNTRACE: back from calling ehChkOvr");

   /*** do some logging ***/

   // puts("EHFUNTRACE::EHFUNTRACE: about to call ehGetFun");
   if ((iGetFun = ehGetFun(iChan)) != 0)
   {
      ehIndent(EHSTDERR);

#ifdef EHSIMCTRLC /* compile-time request for CTRL-C simulation */
      if (Eh.ulCtrlCVal + 1)
         fprintf(EHSTDERR, "%ld ", Eh.ulCtrlCVal + 1);
#endif

      fprintf(EHSTDERR, "%s %s%s%d {",  /* "}" */
            pcFile,
            pcFunc?pcFunc:"", 
            pcFunc?"() ":"", 
            iLine);
      fflush(EHSTDERR);
   }

   /*** CTRL-C longjmp() ***/

   // puts("EHFUNTRACE::EHFUNTRACE: about to call ehChkCtrlC");
   ehChkCtrlC(); /* Eh.iIntSig --> EhSt.{Er.lNo,CtrlC,Chain} */

   /***
      NOTE that the CTRL-C state that we just processed,
      is now considered to be part of the parent's
      environment which, if we longjmp() will become
      part of the catcher's environmemt.
   ***/

   /*** finish logging activity, taking CTRL-C into account ***/

   if ((iInterrupted = INTERRUPTED()) != 0)
   {
      if (iGetFun)
      {
         fputs("   -- interrupted", EHSTDERR);
         if (!just_entering_lev0)
         {
            fputs("\n", EHSTDERR);
            ehIndent(EHSTDERR);                     /* balance out the "{" */
            fprintf(EHSTDERR, "} /""* %s %s%s%d *""/\n",
                  pcFile,
                  pcFunc?pcFunc:"",
                  pcFunc?"() ":"",
                  iLine);
         }
      }
      else /* no function tracing */
      {
         ehIndent(EHSTDERR);
         fprintf(EHSTDERR, "%s %s%s%d -- interrupted\n",
               pcFile,
               pcFunc?pcFunc:"",
               pcFunc?"() ":"",
               iLine);
      }
      ehCallStack(EHSTDERR);
      fflush(EHSTDERR);
   }
   else
   if (iGetFun)
   {
      fputs("\n", EHSTDERR);
      fflush(EHSTDERR);
   }

   /***
      if we have been interrupted, we will try not to execute
      the function we are enterring.  Unless restrictive modes
      are set (yInhibit, yShield), we will longjmp() back to the
      most recent catcher function.  FUN_ENTER() therefore behaves
      as a checkpoint for interrupt (CTRL-C) handling.  If restrictive
      modes are set (such as yInhibit or yShield) we will not longjmp()
      and we will defer CTRL-C handling to FUN_RETURN()/FUN_LEAVE()
      (ehFunRet()) or to the next call to ehFunEnt()
   ***/

   if (iInterrupted) /* CTRL-C + HAVETOJUMP() */
   {
      if (cDmpEhSt)
      {
         fputs("Interrupt detected:", EHSTDERR);
         dmpEhSt(&EhSt);
      }

      EhSt.scError = -1; /* resolve the CTRL-C to a real error */

      if (just_entering_lev0)
      {
         puts("EHFUNTRACE::EHFUNTRACE: patching up EhSt for interrupt in catcher at lev0");
         EhSt.pcFile = pcFile;
         EhSt.pcFunc = pcFunc;
         EhSt.iLine  = iLine;
         ++EhSt.hLev; /* entering next nesting level */
      }

      xhThrow(pcFunc); /* attempt to abort this function and start chaining */
   }

   /***
      we were not interrupted
      or could not chain:
      proceed as normal
   ***/

   /*** Save the current Error Handler state. ***/

   if (!just_entering_lev0) /* save Error Handler state */
   {
      /* save values of our parent */
      memcpy(&ft_EhSt, &EhSt, sizeof(ft_EhSt));
   }

   /***
      Update the Error Handler state.

      If we are being called from main(), this function
      gets called *after* ehCatch(), otherwise it gets
      called *before* ehCatch().  This is determined
      by examining hLev.
   ***/

   /*** establish new state ***/

   EhSt.pEhSv = &ft_EhSt; /* save address of local EHSTATE variable */

   if (!just_entering_lev0)
   {
      EhSt.yCatch = 0;
   }

   ++EhSt.hLev;       /* entering next nesting level */

   EhSt.pcFile = pcFile;
   EhSt.pcFunc = pcFunc;
   EhSt.iLine  = iLine;
   EhSt.yChain = 0;    /* clear the chaining request bit */
   EhSt.yCtrlC = 0;    /* no interrupts have (yet) occurred in this function */
   EhSt.scError = 0;   /* no errors have (yet) occurred in this function */

   if (cDmpEhSt)
   {
      fputs("New Error Handler State:", EHSTDERR);
      dmpEhSt(&EhSt);
   }

   /*** continue first-time initialization, if necessary ***/

   if (Eh.yIniState == 1) // next part, resource initialization
   {
      ++Eh.yIniState; // mainly to prevent recursion

      /* initialize internal resources */
      if (pfvEhIntInit)
      {
         // puts("EHFUNTRACE::ft_PgmEnt: about to call pfvEhIntInit()");
         (*pfvEhIntInit)(); /* initialize internal resources */
      }

      /* initialize external resources */
      if (pfvEhExtInit)
      {
         // puts("EHFUNTRACE::ft_PgmEnt: about to call pfvEhExtInit()");
         (*pfvEhExtInit)(); /* initialize external resources */
      }
   }

   if (cEhFunTrc)
      printf("EHFUNTRACE::EHFUNTRACE: leaving (entering %s, hLev %d)\n",
            pcFunc?pcFunc:"<func>", EhSt.hLev);

} // EHFUNTRACE::EHFUNTRACE


/***
EHFUNTRACE::ft_PgmEnt()
Program-entry tracing support.
***/

void
EHFUNTRACE::ft_PgmEnt(
   int *piArgc,
   char **ppcArgv)
{
   int iRetVal = 0;
   const char *pcEnv;
   char *pc;
   int  iChan = -1;

   if (cEhPgmTrc)
      puts("EHFUNTRACE::ft_PgmEnt: entering");

   cFinished = 0;

   ehCpuChk(); /* test processor characteristics */

   /* open the error log file */
   pcLogFileName = ehOpnLog(Eh.pcProgName ?   Eh.pcProgName : "unknown");

   if (ehInitChnVol(EH_NEWDBGVAL))
   {
      iRetVal = -2;
      goto Exit;
   }

   /* set ulEhCtrlC for SIMCTRLC() */
   // puts("EHFUNTRACE::ft_PgmEnt: about to call getenv(1)");
   Eh.ulCtrlCIni =  ((pc = getenv("EHSIGCTRLC")) != NULL) ?
         atol(pc) : EHCTRLC_DEFAULT ;
   Eh.ulCtrlCVal = Eh.ulCtrlCIni;

   /* check environment variable EHVOLUME */
   if ((pcEnv = getenv("EHVOLUME")) == NULL)
   {
      /* the following means */
      /* -ve set the default volume on all channels to error volume */
      /* -f0 turn function tracing off on all channels */
      /* -m0 turnLeory tracing off on all channels */
      /* -cu the subsequent parameters are for the 'unit" channel */
      /* -vd set the volume on the 'unit' channel to 'debug' level */
      /* -ca the sebsequent parameters are for 'alpha' level */
      /* -vw set the volume on the 'alpha' channel to 'warning' level */
      pcEnv = ""; /* everything is off */
   }

   // set the Error Handler 'volume'
   if ((iRetVal = ehVolume(pcEnv)) != 0)
      goto Exit;

   if (piArgc && ppcArgv)
      ft_Args(piArgc, ppcArgv);

   // set up the dummy save structure
   // puts("EHFUNTRACE::EHFUNTRACE: setting up dummy ft_EhSt");
   ft_EhSt.pcFile   =
   ft_EhSt.pcFunc   = "theAbyss";
   ft_EhSt.iLine    = -1;
   ft_EhSt.pEhSv    = NULL; /* bottom of stack */
   ft_EhSt.hLev     = 0;
   ft_EhSt.yCatch   = 0; /* shouldn't this be here? */
   ft_EhSt.yInhibit = 0;
   ft_EhSt.yShield  = 1; /* never jump out of main */
   ft_EhSt.scError  = 0;
   ft_EhSt.yCtrlC   = 0;
   ft_EhSt.yChain   = 0;
   EhSt.pEhSv = &ft_EhSt;     /* need to do this *before* longjmp() */

#if !ehMX | EH_CHAN
   signal(SIGINT, ehSigInt);
#endif

#if !ehMWCC /* Be Bug */
   // puts("EHFUNTRACE::ft_PgmEnt: about to call atexit()");
   atexit(xhAtExit); /* function to call when user calls exit() */
#endif

   /* announce start time of application */
   // puts("EHFUNTRACE::EHFUNTRACE: about to call ehLogTxt()");
   sprintf(acExecute, "%s\n", Eh.pcProgName ? Eh.pcProgName : "unknown");
   ehLogTxt(acExecute);

   /*** Error Handler is now up ***/

   ttStart = time(NULL); // now we're starting the application

Exit:
   // puts("EHFUNTRACE::ft_PgmEnt: in Exit section");

   /*** if a fatal error has occurred -- no point in continuing ***/

   if (iRetVal < 0)
   {
      fputs("\nCan't continue -- exit()ing...\n", stderr);
      fflush(stderr);
      // puts("EHFUNTRACE::ft_PgmEnt: about to call ehTermChnVol()");
      ehTermChnVol();
      exit(iRetVal);
   }

   if (cEhPgmTrc)
      puts("EHFUNTRACE::ft_PgmEnt: leaving");

   return;
} // EHFUNTRACE::ft_PgmEnt

/***
EHFUNTRACE::Line()
Set the ft_iLine data member.
***/

void
EHFUNTRACE::ft_LineExit(
   const int iLineExit)
{
   //puts("EHFUNTRACE::ft_Line: entering");
   ft_iLineExit = iLineExit;
   //puts("EHFUNTRACE::ft_Line: exiting");
} // EHFUNTRACE::iLine


// EHFUNTRACE Destructor: function un-tracer
EHFUNTRACE::~EHFUNTRACE()
{
   if (cEhFunTrc)
      printf("EHFUNTRACE::~EHFUNTRACE: entering (leaving %s, hLev %d)\n",
            ft_pcFunc?ft_pcFunc:"<func>", EhSt.hLev);

   bool caller_is_catcher = (bool)(EhSt.yCatch != 0);  /* for local use */
   EHSTATE *pEhSvPar = EhSt.pEhSv;        // parent

   /*** if requested, check for memory overwrites ***/

   //puts("EHFUNTRACE::~EHFUNTRACE: about to call ehChkOvr");
   if (ft_chkOvrwrt)
         ehChkOvr(ft_pcFile, ft_pcFunc, ft_iLineExit, &ehIntRes);
   //puts("EHFUNTRACE::~EHFUNTRACE: back from calling ehChkOvr");

   fflush(EHSTDERR);

   /*** if exiting main, begin shutdown procedures ***/

   if (EhSt.hLev == 1 && Eh.yIniState == 2)
   {
      if (cDmpEhSt)
      {
         fputs("shuting down, phase 1:", EHSTDERR);
         dmpEhSt(&EhSt);
      }

      // call all registered termination functions
      ehOsTerm(); /* this is already protected against being called twice */
      Eh.yIniState = 1;
   }

   /*** now do some logging ***/

   //puts("EHFUNTRACE::~EHFUNTRACE: about to call ehLogRet");
   ehLogRet(ft_iChan, ft_pcFile, ft_pcFunc, ft_iLineExit);

   /***
      At this point, we need to examine the current state
      of the error handler and determine whether we are going to:
         o return normally, or
         o longjmp() back to previous catching function.
   ***/ 

   if (cDmpEhSt)
   {
      fputs("Pre-exit state:", EHSTDERR);
      dmpEhSt(&EhSt);
   }

   /***
      yInhibition --
      Much of the time when we call ehFunRet(), we will be in the ExitSequence
      of a function -- either through normal execution or because of an
      exception.  In this case, yInhibition will always be set.  What we want
      is whether our parent is inhibiting us.  Therefore, we get the parent's
      value before testing HAVETOJUMP().

      This is complicated by the occurance of a Ctrl-C during the ExitSequence.
      Because of the inhibition, it will not be processed immediately.  It will
      remain pending until reaching ehFunRet().  If the parent does not have
      inhibition set, we would be tempted to start the chain now.  However, this
      gives the current function no chance to respond to the Ctrl-C.  A resource
      could have been obtained in the current function whose handle is being
      passed back to the parent.  If we process the Ctrl-C here, a normal
      return does not take place, and therefore the parent never sees the
      handle.  However, the current function never learns about the Ctrl-C
      until FUN_RETURN() -- where it is to late to recover.
      
      The parent's inhibition state should be restored *after* the longjmp()
      test is made to ensure that a normal return takes place -- giving resource
      release responsibility to the parent.

      WE MUST set it before checking whether to longjmp() because otherwise we
      can never longjmp from a catcher function: the local ExitSequence will
      have set inhibition.

      WE MUST NOT set it before checking whether to longjmp() because then we
      can loose resource responsibility: suppressed chaining (e.g., from a
      control C) will now become activated and returning a successfully
      obtained resource handle will be impossible.  The resource handle will
      be lost.
   ***/
   EhSt.yInhibit = pEhSvPar->yInhibit; /* possibly turn off inhibition */
   EhSt.yShield  = pEhSvPar->yShield;  /* most restricting condition */
   if (HAVETOJUMP())
   {
      //puts("EHFUNTRACE::~EHFUNTRACE: about to call restoreContext");
      restoreContext(pEhSvPar); /* restore values */
      if ( caller_is_catcher )
      {
         //puts("EHFUNTRACE::~EHFUNTRACE: about to call xhThrow");
         xhThrow(ft_pcFunc); // [re-] throw this exception
      }
      else
         goto Exit;                    // let compiler resume
   }

   /*** de-initialize Error Handler if we are leaving main ***/

   if (EhSt.hLev == 0 && Eh.yIniState == 1)
   {
      /***
         maybe this should be moved to an atexit() function
         because we are only (relatively) sure that the application
         is (really) leaving if we are starting to exit.
         Is there a way to block an atexit?
         Maybe at exit() time, we can set a flag indicating
         a de-initialization request.  As we unwind the stack,
         when we get back to main, we would then really
         de-initialize.  Of course this could be foiled by
         the application choosing to not chain all the way
         out.  Oh no, not another flag to manage!  

         And what happens if we just roll off the end of main?
         I suppose this is the condition that we are trying to
         detect: that is, when we're all chained out and on the
         verge of exiting.  But what about global objects that
         still need to have their destructors called?  Do they
         get called before or after exit()?  If they get called
         *before* exit(), an atexit-based strategy might work.
         If not, then we don't necessarily know *really* exiting,
         or just exiting main().
      ***/
      // puts("EHFUNTRACE::~EHFUNTRACE: about to call ft_PgmExt");
      ft_PgmExt();       // leaving main()
      Eh.yIniState = 0;  // back to your previously scheduled operating system
   }

   /*** normal function return ***/

   /* restore parent's values */
   EhSt.pcFile   = pEhSvPar->pcFile;
   EhSt.pcFunc   = pEhSvPar->pcFunc;
   EhSt.iLine    = pEhSvPar->iLine;
   EhSt.yShield  = pEhSvPar->yShield;  /* possibly turn off shielding */
   EhSt.yCatch   = pEhSvPar->yCatch;
   EhSt.yChain  |= pEhSvPar->yChain;
   EhSt.yCtrlC  |= pEhSvPar->yCtrlC;
   EhSt.hLev     = pEhSvPar->hLev;     /* restore parent's nesting level */
   EhSt.scError |= pEhSvPar->scError;
   EhSt.pEhSv    = pEhSvPar->pEhSv;

Exit:
   if (cEhFunTrc)
      printf("EHFUNTRACE::~EHFUNTRACE: leaving (leaving %s, hLev %d)\n",
            ft_pcFunc?ft_pcFunc:"<func>", EhSt.hLev);
} // EHFUNTRACE::~EHFUNTRACE


/***
EHFUNTRACE::ft_PgmExt()
Program-exit tracing support.
***/

void
EHFUNTRACE::ft_PgmExt()
{
   time_t ttElapsed;

   if (cEhPgmTrc)
      puts("EHFUNTRACE::ft_PgmExt: entering");

   //puts("EHFUNTRACE::ft_PgmExt: calculating ttElapsed");
   ttElapsed = time(NULL) - ttStart;
   //puts("EHFUNTRACE::ft_PgmExt: calling sprintf");
   sprintf(acExecute, "%s: %ld sec%s\n",
         Eh.pcProgName ? Eh.pcProgName : "unknown",
         (long)ttElapsed, ttElapsed != 1 ? "s" : "");
   //puts("EHFUNTRACE::ft_PgmExt: calling ehLogTxt");
   ehLogTxt(acExecute);

   //printf("ft_PgmExt: Eh.Er.acPrefix = %s\n", Eh.Er.acPrefix);
   //printf("ft_PgmExt: Eh.Er.u.hCode  = %d\n", Eh.Er.u.hCode);

   if (Eh.Er.hCode)
   {
      const char *pc = "process failed with error %s%c%04x";
      const char c = EhSt.scError < 0 ? '-' : '+';

      if (Eh.Er.hCode < 0)
      {
         Eh.Er.hCode = (short)-Eh.Er.hCode;
      }
      fprintf(EHSTDERR, pc, Eh.Er.acPrefix, c, Eh.Er.hCode);
      fprintf(EHSTDERR, "\n\n");
      fflush(EHSTDERR);
      /* if error stream is file, write also to console */
      if (ehGetOut() != 1)
      {
         printf(pc, Eh.Er.acPrefix, c, Eh.Er.hCode);
         if (pcLogFileName)
            printf(", see error log file %s", pcLogFileName);
         printf("\n\n");
      }
   }
   if (EhSt.yCtrlC)
   {
      const char *pc = "Control-C (interrupt signal) detected\n\n";
      fprintf(EHSTDERR, pc);
      fflush(EHSTDERR);
   }

   //puts("EHFUNTRACE::ft_PgmExt: about to call ehTermChnVol");
   ehTermChnVol();
   fflush(EHSTDERR);

   //puts("EHFUNTRACE::ft_PgmExt: about to call ehClsLog");
   ehClsLog(); /* NOTHING must write to stderr after this line !!! */
               /* be very careful about the ehAtExit() function    */

   cFinished = 1;

   if (cEhPgmTrc)
      puts("EHFUNTRACE::ft_PgmExt: leaving");
} // EHFUNTRACE::ft_PgmExt


/***
EHFUNTRACE::Catch()
This function enables catching.  It is called via the FUN_CATCH()
macro.  If FUN_CATCH() is not called, then a function is not
capable of receiving a thrown exception.  This has the same effect
as being shielded.  In other words, if FUN_CATCH() is not called,
and a child function wants to throw an exception, it won't be able
to.  This has the same effect as whenever we cannot throw an exception:
further chaining is stopped.
***/

void
EHFUNTRACE::ft_Catch()
{
   /* not yet incremented by ehFunEnt() */
   bool just_entering_lev0 = (bool)(EhSt.hLev == 0);

   // puts("EHFUNTRACE::Catch: entering");

   /*** we are a catcher ***/

   EhSt.yCatch  = 1; /* we are a catcher */
   EhSt.yShield = 0; /* let us catch! */

   /***
      Update the Error Handler state.

      If we are being called from main(), this function
      gets called *before* ehFunEnt(), otherwise it gets
      called *after* ehFunEnt().  This is determined
      by examining hLev.
   ***/

   //if (!just_entering_lev0) /* setjmp() has been called at *least* once */
      //memcpy(ft_EhSt.JmpBuf, EhSt.JmpBuf, sizeof(ft_EhSt.JmpBuf));

   // puts("EHFUNTRACE::Catch: leaving");

} // EHFUNTRACE::ft_Catch()


/***
restoreContext()
This function assumes that it is OK to longjmp().
No checks are made.  It sets several state
variable according to the parent's and/or
catcher's values.  Then performs the longjmp().
??? Should shielding be reset here ???
Shielding does not prevent jumping from a
shielded function, only jumping to a 
shielded function.  Therefore, we may be
in a shielded function jumping to an un-
shielded function, in which case we should
reset yShield according to our destination.
***/

static
void
restoreContext(
   EHSTATE*pEhSvPar) /* Parent */
{
   // const char *pcFunc = EhSt.pcFunc; // remember this before changing

   if (EhSt.yCtrlC)
      EhSt.scError = -1; /* resolve the Control-C to an error */

   EhSt.pcFile   = pEhSvPar->pcFile;
   EhSt.pcFunc   = pEhSvPar->pcFunc;
   EhSt.iLine    = pEhSvPar->iLine;
   EhSt.yShield  = pEhSvPar->yShield; /* ??? */
   EhSt.yCatch   = pEhSvPar->yCatch;
   EhSt.yCtrlC  |= pEhSvPar->yCtrlC;  /* pick worst condition */
   EhSt.hLev     = pEhSvPar->hLev;                         /* restore catcher's nesting level */
   EhSt.scError |= pEhSvPar->scError; /* pick most restricting condition */
   EhSt.pEhSv    = pEhSvPar->pEhSv;                        /* restore catcher's EHSTATE variable */
} // restoreContext()


/***
ehPErrorThrow()
This function is called from
   ehPError() - in XHTHROW_XXX()
This function executs a throw to the nearest ExitSequence.
This ExitSequence may be in the current function
(if the current function is a catcher) or somewhere
up the call stack.  If the current function is a
catcher, we can ignore inhibition and shielding and
simply thow to the current ExitSequence.  If the current
function is not a catcher, we may be inhibited or
shielded by a functino further up the call stack.
***/

void
ehPErrorThrow(
   const char *pcFunc)
{
   EHSTATE *pEhSv = &EhSt;         /* point to current function */
   EHSTATE *pEhSvPar = EhSt.pEhSv; /* point to parent */
   EHSTATE *pEhSvCat = pEhSv;      /* point to most recent catcher */
   byte yShield;     /* most restricting condition */

   if (cDmpEhSt) dmpEhSt(&EhSt);

   /* find the EHSTATE of the most recent catcher */
   while (pEhSvCat->pEhSv && pEhSvCat->yCatch == 0)  /* find nearest catcher */
      pEhSvCat = pEhSvCat->pEhSv;

   /*** can we just goto a local ExitSequence? ***/

   if (pEhSv == pEhSvCat) /* current function is a catcher */
   {
      //puts("ehPErrorThrow: about to call xhThrow (1)");
      xhThrow(pcFunc); /* makes debugging easier: stops in xhCaught() */
      //return;          /* do local goto ExitSequence */
   }

   /*** if allowed, jump to catcher ***/

   /* see if we are allowed to longjmp() */
   yShield  = (byte)(pEhSv->yShield || pEhSvPar->yShield || pEhSvCat->yShield);
   if (EhSt.yChain && !EhSt.yInhibit && !yShield)
   {
      //puts("ehPErrorThrow: about to call xhThrow (2)");
      xhThrow(pcFunc);  // restoration will take place in EHFUNTRACE::~EHFUNTRACE momentarily
   }
} /* xhPerrorThrow */


/***
xhThrow()
Logs and performs a longjmp()...no checking.
***/

static
void
xhThrow(
   const char *pcFunc) /* may be NULL */
{
   if (cDmpEhSt)
   {
      fputs("about to throw:", EHSTDERR);
      dmpEhSt(&EhSt);
   }

   ehIndent(EHSTDERR);
   fputs("*** thrown", EHSTDERR);
   if (pcFunc) fprintf(EHSTDERR, " from %s()", pcFunc);
   fputs(" *** \n", EHSTDERR);
   fflush(EHSTDERR);

   throw Eh.Er; /*** !!! >>> throw <<< !!! ***/
} // xhThrow()


/***
xhCaught()
This function gets called after the throw
and after any stack unwinding has taken place
(e.g., the EHFUNTRACE destructor is called).
***/

void
xhCaught A1(const char *, pcFunc)
{
   ehIndent(EHSTDERR);
   fprintf(EHSTDERR, "*** caught in ");
   if (pcFunc) fprintf(EHSTDERR, "%s()'s ", pcFunc);
   fprintf(EHSTDERR, "Cleanup Section ***", pcFunc);
   if (cDmpEhSt) dmpEhSt(&EhSt);
   fprintf(EHSTDERR, "\n");
   fflush(EHSTDERR);
} /* xhCaught */

/***
xhAtExit()
C library calls this function when exit() is called.
***/

#define TRACE_ATEXIT 0

static
void
xhAtExit()
{
#if TRACE_ATEXIT
   ehIndent(stdout);
   fputs("--- entering ehAtExit() ---\n", stdout);
#endif

   if (!cFinished)
   {
      EhSt.yChain = 1;   /* request chaining! */
      xhExitThrow();     /* try to jump */

      ehOsTerm(); /* this is already protected against being called twice */
      ehTermChnVol();
      ehClsLog(); /* NOTHING must write to stderr after this line !!! */
   }

#if TRACE_ATEXIT
   ehIndent(stdout);
   fputs("--- exiting ehAtExit() ---\n", stdout);
#endif
} // xhAtExit()

/***
xhExitJmp()
This function is called if user calls exit().
First xhAtExit() is called which determines
whether chaining is appropriate.  If so, it
calls this function to attempt to initiate
chaining.  If we are not allowed to chain,
we can just return, at which point xhAtExit()
will do system cleanup.  Otherwise, we can
commense chaining and will eventually end up
back in main() where final system cleanup
will be performed.
***/

static
void
xhExitThrow()
{
   EHSTATE *pEhSvPar = EhSt.pEhSv; /* point to parent */

   if (cDmpEhSt) dmpEhSt(&EhSt);

   /*** may we start chaining ? ***/

   /* this is a last-ditch attempt */
   if (!EhSt.yInhibit && !EhSt.yShield)
   {
      puts("xhExitThrow: about to call xhThrow");
      xhThrow(EhSt.pcFunc); // [re-] throw this exception
   }
} /* xhExitThrow() */

#endif /* ehEXCEPTIONHANDLING */
