/***

ehintres.h -- Error Handler, internal resource tracking

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

#ifndef EHINTRES_H
#define EHINTRES_H 1
static const char*ehintres =
       "$Id: ehintres.h,v 1.5 1996/01/09 02:31:41 ccobb Exp $";

#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE 1
#endif

#ifndef   EH_H
#include "eh.h"
#endif

/***

The Error Handler has internal support for tracking the
use of a set of resources (memory, file pointers and
file descriptors).  It can also support a parallel set
of externally defined resources.

This header is part of the internal implementation of 
tracking a set of resources.  External implementation
can be set up in an indentical fashion.

This is the top level header file which identifies all
of the resources types to be tracked internally via an
enumerated type.  The resources tracked internally are:
dynamic memory, file pointers and file descriptors.
These resources are tracked using a single data structure
(EHRESTRK, defined in eh.h).  Function prototypes are defined
here (see bottom of this file) for the functions
that initialize and de-initialize (terminate) use of
this data structure:
   void ehIntResInit(void);
   void ehIntResTerm(void);
These functions take no arguments and refer internally
to an instance of an EHRESTRK structure named ehIntRes.  

The Error Handler's internal resource initialization 
function, ehIntResInit(), is called by ehIntIni(),
the general internal initialization function.

To track external resources, start with a copy of this
header and change the names to reflect the the resources
you would like to track externally.  A similar pair of
initialization and termination functions would be necessary
which make reference to a different instance of a EHRESTRK
structure, e.g., myExtRes.  Your functions might be called:
   void myExtResInit(void);
   void myExtResTerm(void);
Similarly, the external resource initialization function would 
probably be called by a general initialization function which
might be called myExtIni().

There is one final twist which completes the initialization loop.
The Error Handler needs to be able to call the external initialization
function at start up -- but only if one is present.  This is 
accomplished by calling this function indirectly through a function
pointer.  This function pointer is set to null in the Error
Handler library.  If the pointer's value is null, no external
initialization function is called.  But an application can
define a non-null instance of this pointer which would pre-empt
the built-in instance.  In this case, the external initialization
function which is pointed to by this pointer will be called at
startup time.  This pointer has the name pfvEhExtInit and is
declared in eh.h.

For purposes of consistency and example, the internal initialization
process is set up in exactly the same way.  A function pointer
named pfvEhIntInit is declared in eh.h.  At program start up time,
this pointer is check to see if it has a non-null value (which it
always will have), and if so, it calls the internal initialization
function to which it points.

This function pointer is defined in ehintres.c, which is the
sibling .c file which corresponds to the header file you are
now looking at.  You will want to have a file, myextres.c, 
which will contain the definitions for the external resource
tracking functions that you will define, including pfvEhExtInit.

The user (programmer) of the internal resource tracking
system sees the following header files:

   ehmemtrk.h -- for memory tracking
   ehfiltrk.h -- for FILE pointer tracking
   ehfdtrk.h  -- for file descriptor tracking

Even though the same (single) set of data structures are
used to track all three of these resources, separate header
files are supplied to give the user fine grained control
over the resources he wants to track.  These headers define
the wrapper macros which intercept the resource acquisition
function calls.  The user would include one or more of these
header files in his application as desired.

An alternate set of header files can also be used:

   ehmem.h -- for memory tracking
   ehfil.h -- for FILE pointer tracking
   ehfd.h  -- for file descriptor tracking

These header files are needed when the user has not yet
implemented function tracing in his source code.  These
cannot be intermixed with any of the previous three in
the same source file.  This means that when initially
implementing resource tracking, function tracing must
be implemented in either all of the functions in the
source file or none of them.

A set of associated .c files implement the wrapper functions
which implement the resource tracking.  For the internal
resource tracking these are:

   ehmemtrk.c -- for memory tracking
   ehfiltrk.c -- for FILE pointer tracking
   ehfdtrk.c  -- for file descriptor tracking

***/

/***
   This is the top-level structure through which all internal
   resource tracking is performed.
***/
extern EHRESTRK ehIntRes; /* internal resource tracking anchor */

/*** Internal Resources ***/

/* internally tracked resources */
typedef enum _eh_int_res_
{
   EH_RES_UNKNOWN,  /* un-determined (yet) */
   EH_RES_HEAPMEM,  /* dynamically allocated memory */
   EH_RES_FILESTR,  /* FILE stream pointer */
   EH_RES_FILEDES   /* file descriptor */
} EHINTRES;         /* internally tracked resource types */

void ehIntResInit(void); /* initialize internal resource tracking */
void ehIntResTerm(void); /* terminate  internal resource tracking */

#endif /* EHINTRES_H */

