/***

ehtrkres.c -- Error Handler, general-purpose resource tracking

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char*rcsid =
       "$Id: ehtrkres.c,v 1.10 1996/03/05 22:10:58 ccobb Exp $";

/*** standard headers ***/

#include "ehenv.h"

EHBEGINEXTERNC
#include <string.h>
#include <stdlib.h>
EHENDEXTERNC

/*** amiga headers ***/

#define EH_NTRKRES 1 /* don't track resources within this file! */

/*** error handler headers ***/

#ifndef EH_CHAN
#define EH_CHAN 1
#endif
#include "eh.h"
#include "eherrs.h"

#include "ehavltre.h"
#include "ehresini.h"

#define EH_NMEMTRK 1
#include "ehmemtrk.h"

#if     TRACKING_INTERNAL_MEMORY_USE
#include <mt.h>
#endif

/*** local structures ***/

typedef struct /* function tracing support */
{
   const char *pcFile; /* file name */
   const char *pcFunc; /* function name */
   int   iLine;        /* line number of function call */
   ulong ulCount;      /* # times this line has been executed/passed */
} EHCALL;

/*** local headers ***/

static EHRESLIST*insert_new_node A((EHCALL*, EHRESTRK*, EHRESLIST*, ulong**));
static void      update_dup_node A((EHCALL*, EHRESTRK*, EHRESLIST*, ulong **));
static EHRESLIST*compare_handles A((EHCALL*, EHRESTRK*, EHRESLIST*, EHRESLIST*,
	   ulong**));
static EHRESLIST*find_old_node   A((EHRESTRK*, EHRESLIST*, ulong));
static EHRESLIST*update_old_node A((EHCALL*, EHRESTRK*, EHRESLIST*, ulong**));
static void      reuse_old_node  A((EHCALL*, EHRESTRK*, EHRESLIST*, ulong**));
static ulong   update_pass_count A((EHRESTRK *, int iLine, const char *pcFile));

EHFILENAME(ehrestrk.c);

/***
ehResObt()
Track resource usage.
This function is a wrapper to a resource-obtaining function.  This function
calls the original resource-obtaining function, but also keeps track of
what resources are in use.

Error Handling (forward references)
(1)  update_pass_count() fails
(2)  EHASSERT_EXITSEQ(iNumArgs > 1)
(3)  EHASSERT_EXITSEQ(iNumArgs > 1)
(4)  EHTHROW(EH_CALLOC) pulArgs
[a]  pulArgs is obtained
(5)  EHASSERT_EXITSEQ(pcResName != NULL) fails
(6)  EHTHROW(EH_CALLOC) pcResource
[b]  pcResource is obtained
(7)  ehAvlSrch(pEhResTrk->pEhAvlRes, (const void*)&ResList) fails
(8)  insert_new_node(&EhCall, pEhResTrk, &ResList,...) fails
[c]  new node is inserted
(9)  update_dup_node(&EhCall,...) fails
[d]  existing node is updated
(10) reuse_old_node(&EhCall, pEhResTrk, pEhResList, ...) fails
[e]  old node is reused

Error Handling (reverse references)
(10) reuse_old_node(&EhCall, pEhResTrk, pEhResList, ...) fails
      [b]  pcResource was obtained -- free it
      [a]  pulArgs was obtained -- free it
(9)  update_dup_node(&EhCall,...) fails
      [b]  pcResource was obtained -- free it
      [a]  pulArgs was obtained -- free it
(8)  insert_new_node(&EhCall, pEhResTrk, &ResList,...) fails
      [b]  pcResource was obtained -- free it
      [a]  pulArgs was obtained -- free it
(7)  ehAvlSrch(pEhResTrk->pEhAvlRes, (const void*)&ResList) fails
      [b]  pcResource was obtained -- free it
      [a]  pulArgs was obtained -- free it
(6)  EHTHROW(EH_CALLOC) pcResource
      [a]  pulArgs was obtained -- free it
(5)  EHASSERT_EXITSEQ(pcResName != NULL) fails
      [a]  pulArgs was obtained -- free it

Error Handling (summary)
   [b]  if pcResource was obtained -- free it
   [a]  if pulArgs was obtained -- free it
***/

bool
ehResObt A8V(
   const char *, pcFile,
   const char *, pcFunc,
   int,          iLine,
   EHRESTRK *,   pEhResTrk,   /* resource tracking header */
   int,          iType,
   const char *, pcResName,   /* text name for resource */
   pfv_ulv,      pfvRelease,  /* ptr func ret void */
   ulong,        ulFlags,
   ...)
{
   /* used in Clean Up section */
   ulong *volatile pulArgs = NULL;
   char *volatile pcResource = NULL;
   EHRESLIST *volatile pEhResList = NULL;
   volatile int iNumArgs = -1;
   ulong aulArgs[EH_RES_NUMARGS_MSK]; /* for error recovery */
   EHRESLIST ResList;

   EHCATCH(ehResObt) {

      /* not used in Clean Up section */
      volatile va_list ap;
      EHCALL EhCall;
      int iArg;

      if (cEhResDmp)
      {
         fprintf(EHSTDERR, "ehResObt(%s, %s, %d, %d, 0x%lx, ...)\n", 
               pcFile, pcFunc?pcFunc:"<func>", iLine, iType, ulFlags);
         fflush(EHSTDERR);
         ehAvlTreeDump(pEhResTrk->pEhAvlRes);
      }

      /*** initialize structures ***/

      EhCall.pcFile = pcFile;
      EhCall.pcFunc = pcFunc;
      EhCall.iLine  = iLine;
      EhCall.ulCount = update_pass_count(pEhResTrk, iLine, pcFile);
      //EHERR1("EhCall.ulCount = %lu", EhCall.ulCount);

      /***
         In the case of ehResObt() we can pre-initialize a local copy
         of ResList which will get duplicated as necessary and will
         already contain the correct values, e.g., pfvDefRel.   In
         other scenarios, we may be working with new or previously
         existing buffers, so these cannot be initialize/updated
         until we have a pointer to the target EHRESLIST buffer.
      ***/

      memset(&ResList, 0, sizeof(ResList));
      ResList.ulFlags = ulFlags;
      ResList.pfvDefRel = pfvRelease;
      ResList.EhResState = EH_RES_RELEASED; /* so far */
      ResList.EhResWhom = EH_RES_NONE;      /* initialize */

      /*** get info needed for recovery ***/

      /* extract count from flags */
      iNumArgs = (int)(ulFlags & EH_RES_NUMARGS_MSK);

      /* populate local array */
      /* load array with arguments */
      /* this copy is for error recovery */
      VA_START(ap, ulFlags); 
      for (iArg = 0; iArg < iNumArgs; ++iArg)
      {
         aulArgs[iArg] = va_arg(ap, ulong);
      }
      va_end(ap);

      /* handle functions which don't return a value */
      if (ulFlags & EH_RES_VOID)
      {
         EHASSERT_EXITSEQ(iNumArgs == 0);
         ResList.ulResHndl = EH_RES_VOID; /* arbitrary tag for void resources */
      }
      else if (ulFlags & EH_RES_HNDL_2ND)
      {
         EHASSERT_EXITSEQ(iNumArgs > 1);
         ResList.ulResHndl = aulArgs[1];
      }
      else /* switch on # args and use appropriate function call */
      {
         ResList.ulResHndl = aulArgs[0];
      }
      ResList.iType = iType;

      /***
         at this point we have enough info
         to recover from an error, so now
         we can call failable functions.
      ***/

      /*** read off variable args into dynamic array ***/

      /* create temp array of count longs */
      if (iNumArgs)
      {
         /* this copy is for normal use.. */
         /* it will get attached to EHRESCALL structure */
         if ((pulArgs = (ulong*)calloc(sizeof(ulong), iNumArgs)) == NULL)
            EHTHROW(EH_CALLOC);

         /* load array with arguments */
         for (iArg = 0; iArg < iNumArgs; ++iArg)
         {
            pulArgs[iArg] = aulArgs[iArg];
         }
      }

      /*** make local copy of resource name ***/

      EHASSERT_EXITSEQ(pcResName != NULL);
      if ((pcResource = (char *)calloc(strlen(pcResName)+1, 1)) == NULL) 
               EHTHROW(EH_CALLOC);
      strcpy(pcResource, pcResName);

      /*** has this been previously obtained ? ***/

      /* lookup resource handle */
      EHREM1("resource handle is 0x%lx", ResList.ulResHndl);
      if (pEhResTrk->pEhAvlRes != NULL)
      {
         EHREM("is resource in table?");
         if (cEhAvlDmp) ehAvlTreeDump(pEhResTrk->pEhAvlRes);
         pEhResList =  (EHRESLIST *)
                   ehAvlSrch(pEhResTrk->pEhAvlRes, (const void*)&ResList);
      }

      if (pEhResList == NULL) /* resource handle not in tree */
      {
         pEhResList = insert_new_node(&EhCall, pEhResTrk, &ResList,
                             (ulong **)&pulArgs);
         pEhResList->pcResNam = pcResource;
         pcResource = NULL;
      }
      else /* resource handle already in table */
      {
         if (pEhResList->EhResState == EH_RES_OBTAINED &&
                       (ulFlags & EH_RES_ALLOW_DUPS))
         {
            /* user can allocate same resource multiple times */
            update_dup_node(&EhCall,
                  pEhResTrk, pEhResList, (ulong **)&pulArgs);
         }
         else /* same resource cannot be obtained more than once */
         {
            /* new resource is reusing old node */
            reuse_old_node(&EhCall, pEhResTrk, pEhResList,
                  (ulong **)&pulArgs);
            pEhResList->pcResNam = pcResource;
            pcResource = NULL;
         }
         pEhResList->EhResWhom  = EH_RES_OBTAIN;
         pEhResList->EhResState = EH_RES_OBTAINED;
         pEhResList->pfvDefRel  = pfvRelease; /* set default release function */
      }

      ++pEhResTrk->ulResObtained;
      pEhResTrk->ulResPeak = MAX(pEhResTrk->ulResPeak, 
                           pEhResTrk->ulResObtained - pEhResTrk->ulResReleased);

   } EHENDCATCH;

   if (cEhAvlDmp) ehAvlTreeDump(pEhResTrk->pEhAvlRes);
   if (cEhResDmp) ehResTreeDump(pEhResTrk->pEhAvlRes);

   if (ehError())
   {
      /* resource was obtained, release it */
      if (ResList.ulResHndl)
      {
         ResList.ulResHndl = 0L;
      }
   }
   if (pcResource) free(pcResource);
   if (pulArgs) free(pulArgs);
   EHRETURN(!ehError()); /* completed successfully: not interrupted */
} /* ehResObt */


/***
insert_new_node()
Inserts new resource tracking node into the AVL tree.

Error Handling: (failure points) [resources]
(1) EHTHROW(EH_NEWMACRO);
[a] pEhResList allocated
(2) EHTHROW(EH_NEWMACRO);
[b] pEhResCall allocated
(3) ehAddHead() fails
[c] node added to list
(4) ehAvlAdd() fails
[d] szResCount incremented

Error Handling, reverse references:
(4) ehAvlAdd() fails
    [c] node added to list -- need to remove from list
    [b] pEhResCall was allocated -- free pEhResCall
    [a] pEhResList was allocated -- free pEhResList
(3) ehAddHead() fails
    [b] pEhResCall was allocated -- free it
    [a] pEhResList was allocated -- free it
(2) EHTHROW(EH_NEWMACRO);
    [a] pEhResList was allocated -- free it

Error Handling, summary:
   [c] if node was added to list, remove it
   [b] if pEhResCall was allocated, free it
   [a] if pEhResList was allocated, free it
***/

static
EHRESLIST *
insert_new_node A4(
   EHCALL *,     pEhCall,
   EHRESTRK *,   pEhResTrk,
   EHRESLIST *,  pEhResListOrg, /* ulResHndl, iType should be set */
   ulong **,     ppulArgs)
{
   EHRESCALL *volatile pEhResCall = NULL;
   EHRESLIST *volatile pEhResList = NULL;
   volatile bool headAdded = false;

   EHCATCH(insert_new_node) {

      //EHDBGX(ehResTreeDump(pEhResTrk->pEhAvlRes));
      EHREM1("--- inserting new node H:%lx ---", pEhResListOrg->ulResHndl);

      /*** never been used before: insert new record ***/

      /* make dynamic copy of reslist */
      EHDBG("about to NEW(EHRESLIST)");
      if ((pEhResList = NEW(EHRESLIST)) == NULL)
         EHTHROW(EH_NEWMACRO);
      EHDBG("about to memcpy");
      EHDBG3("pEhResList = %p, pEhResListOrg = %p, sizeof(*pEhResLlist) = %d",
            pEhResList, pEhResListOrg, sizeof(*pEhResList));
      EHDBGX(ehCallStack(EHSTDERR));
      memcpy(pEhResList, pEhResListOrg, sizeof(*pEhResList));

      /* update tracking info */
      EHDBG("updating tracking info");
      pEhResList->EhResWhom  = EH_RES_OBTAIN;
      pEhResList->EhResState = EH_RES_OBTAINED;
      pEhResList->hasProblem = false;

      EHDBG("about to NEW(EHRESCALL)");
      if ((pEhResCall = pEhResList->pEhResCall = NEW(EHRESCALL)) == NULL)
         EHTHROW(EH_NEWMACRO);

      EHDBG("updating more tracking info");
      pEhResList->uhNumObtn = (ushort)1;
      pEhResList->szRCInUse = pEhResList->szRCAlloc = (size_t)1;

      pEhResCall->pulArgs = *ppulArgs;
      *ppulArgs = NULL;
      pEhResCall->pcFileO  = pEhCall->pcFile;
      pEhResCall->pcFuncO  = pEhCall->pcFunc;
      pEhResCall->iLineO   = pEhCall->iLine;
      pEhResCall->ulCountO = pEhCall->ulCount;
      pEhResCall->pcFileR = NULL;
      pEhResCall->pcFuncR = NULL;
      pEhResCall->iLineR = 0;
      pEhResCall->ulCountR = 0L;
      pEhResCall->pEhResList = pEhResList; /* pointer back to parent */

      /* insert new ResCall into linked list */
      EHDBG("about to ehAddHead()");
      headAdded = ehAddHead(&(pEhResTrk->EhList), &(pEhResCall->EhNode));

      /* insert new ResList into tree */
      EHDBG("about to ehAvlAdd()");
      ehAvlAdd(
            pEhResTrk->pEhAvlRes,
            (void *)pEhResList,
            (pfi_pvpv)NULL);
      ++pEhResTrk->szResCount;

   } EHENDCATCH;

   EHDBG("in Clean Up section");
   if (ehError())
   {
      EHERR("--- Resrc Insertion Failed ---");
      if (headAdded)
      {
         ehRemHead(&(pEhResTrk->EhList)); /* remove node from list */
      }
      /* a new EHRESCALL record was allocated, free it */
      if (pEhResCall)
      {
         if (pEhResCall->pulArgs)
         {
            *ppulArgs = pEhResCall->pulArgs; /* needed upstairs */
            pEhResCall->pulArgs = NULL;
         }
         EHFREE(pEhResCall);
      }
      if (pEhResList)
      {
         if (pEhResList->pcResNam)
         {
            free( (char *)(((EHRESLIST *)pEhResList)->pcResNam) );
            pEhResList->pcResNam = NULL;
         }
         EHFREE(pEhResList);
      }
   }
   else
   {
      EHREM("--- New Resource Inserted ---");
   }
   if (pEhResList)
   {
      if (pEhResList->pcResNam) EHREM1("pEhResList->pcResNam: %s",pEhResList->pcResNam);
      EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
      EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
      EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);
   }
   if (pEhResCall) EHREM1("pEhResCall->pulArgs: %p",  pEhResCall->pulArgs);

   EHDBGX(ehCallStack(EHSTDERR));
   EHDBG("about to call EHLEAVE()");
   EHRETURN(pEhResList);
} /* insert_new_node */


/***
update_dup_node()
Update an existing node for a duplicate resource
which has already been obtained.

Error Handling: (failure points) [resources]
(1) EHTHROW(EH_CALLOC);
[a] ppulArray is allocated
(2) EHTHROW(EH_DYNARR);

Error Handling: reverse references
(2) EHTHROW(EH_DYNARR);
      [a] ppulArray was allocated -- free it

Error Handling: summary
   [a] if ppulArray was allocated -- free it
***/

static
void
update_dup_node A4(
   EHCALL *,     pEhCall,
   EHRESTRK *,   pEhResTrk,
   EHRESLIST *,  pEhResList,
   ulong **,     ppulArgs)
{
   EHRESCALL *volatile pEhResCall = NULL;
   unsigned long **volatile ppulArray = NULL;
   volatile size_t szArray = 0;
   unsigned long **ppul;
   int i;

   EHCATCH(update_dup_node) {

      EHREM("--- DUP: resource was previously OBTAINED ---");
      if (pEhResList->pcResNam)
         EHREM1("pEhResList->pcResNam: %s",pEhResList->pcResNam);
      EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
      EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
      EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);

      /***
      ok, this is a little bit excessive, but it seals
      up a possible memory leak.  The problem is that 
      argument arrays (pulArgs) are dynamically allocated and kept
      within the ResCall structure which is itself dynamically
      allocated.  If the reallocation (using ehDynArr()) of 
      the ResCall structure fails, then we lose our pointers
      to our pulArgs.  In order to be able to free them properly
      in case of an error, we need to store them somewhere.  So
      we will dynamically allocate (!) a new structure to remember
      our pointers just so we can free them properly in case
      ehDynArr() fails.  Is this convoluted enough for you???
      ***/

      szArray = pEhResList->szRCInUse;
      ppulArray = (ulong**)calloc(szArray, sizeof(long *));
      if (ppulArray == NULL) EHTHROW(EH_CALLOC);

      ppul = ppulArray;
      i = (int)szArray;
      pEhResCall = pEhResList->pEhResCall;
      while (i--)
      {
         *ppul++ = pEhResCall->pulArgs;
         ++pEhResCall;
      }
      pEhResCall = NULL;

      if (pEhResList->szRCInUse + 1 > pEhResList->szRCAlloc)
      {
         /* extend the resource call list for one additional function */
         pEhResCall = (EHRESCALL*)ehDynArr((void**)&(pEhResList->pEhResCall), 
	  &(pEhResList->szRCAlloc),
	  sizeof(EHRESCALL));
         if (pEhResCall == NULL) EHTHROW(EH_DYNARR);
      }
      else
      {
         pEhResCall = pEhResList->pEhResCall + pEhResList->szRCInUse;
      }
      ++pEhResList->szRCInUse;
      ++pEhResList->uhNumObtn;

      /* set values for instance of resource function call */
      pEhResCall->pulArgs = *ppulArgs;
      *ppulArgs = NULL;
      pEhResCall->pcFileO  = pEhCall->pcFile;
      pEhResCall->pcFuncO  = pEhCall->pcFunc;
      pEhResCall->iLineO   = pEhCall->iLine;
      pEhResCall->ulCountO = pEhCall->ulCount;
      pEhResCall->pcFileR = NULL;
      pEhResCall->pcFuncR = NULL;
      pEhResCall->iLineR = 0;
      pEhResCall->ulCountR = 0L;

   } EHENDCATCH;

   if (ehError())
   {
      if (ppulArray != NULL) /* ehDynArr() failed */
      {
         /* free our dangling memory */
         ppul = ppulArray;
         i = (int)szArray;
         while (i--)
         {
            if (ppul) free(ppul);
            ++ppul;
         }
      }
   }
   else
   {
      EHREM("--- resource updated ---");
   }
   if (pEhResList->pcResNam) EHREM1("pEhResList->pcResNam: %s",pEhResList->pcResNam);
   EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
   EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
   EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);
   if (pEhResCall) EHREM1("pEhResCall->pulArgs: %p",  pEhResCall->pulArgs);

   if (ppulArray) free(ppulArray);
   EHLEAVE();
} /* update_dup_node */


/***
update_old_node()
Update an existing node representing same active resource
with new tracking info.  This is currently only used when
re-obtaining a resource.

Error Handling: (failure points) [resources]
(1) EHASSERT_EXITSEQ(pEhResList->EhResState == EH_RES_OBTAINED);
    EHASSERT_EXITSEQ(pEhResList->szRCAlloc >= 1);
    EHASSERT_EXITSEQ(pEhResList->pEhResCall != NULL);
[a] pcResource is transfered to pEhResList->pcResNam
[b] pResList is updated
[c] pResCall is updated
(2) ehAddHead() failed

Error Handling: reverse references
(2) ehAddHead() failed
      [c] pResCall was updated -- restore it
      [b] pResList was updated -- restore it
      [a] pcResource was transfered to pEhResList->pcResNam -- reverse it?

Error Handling: summary
   [c] if pResCall was updated -- restore it
   [b] if pResList was updated -- restore it
   [a] if pcResource was transfered to pEhResList->pcResNam -- reverse it
***/

static
EHRESLIST*
update_old_node A4(
   EHCALL *,     pEhCall,
   EHRESTRK *,   pEhResTrk,
   EHRESLIST *,  pEhResList,
   ulong **,     ppulArgs)
{
   EHRESCALL *volatile pEhResCall = NULL;

   EHCATCH(update_old_node) {

      /* it must be OBTAINED */
      EHASSERT_EXITSEQ(pEhResList->EhResState == EH_RES_OBTAINED);
      /* there must exist a EHRESCALL rec */
      EHASSERT_EXITSEQ(pEhResList->pEhResCall != NULL);
      /* there must be exactly one EHRESCALL rec */
      EHASSERT_EXITSEQ(pEhResList->szRCAlloc == 1);

      EHREM("--- updating old node ---");
      EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
      EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
      EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);

      /* update tracking info */
      pEhResList->EhResWhom = EH_RES_REOBTAIN;
      pEhResCall = pEhResList->pEhResCall;

      /* set values for instance of resource function call */
      if (pEhResCall->pulArgs) EHFREE(pEhResCall->pulArgs);
      pEhResCall->pulArgs = *ppulArgs;
      *ppulArgs = NULL;
      pEhResCall->pcFileO  = pEhCall->pcFile;
      pEhResCall->pcFuncO  = pEhCall->pcFunc;
      pEhResCall->iLineO   = pEhCall->iLine;
      pEhResCall->ulCountO = pEhCall->ulCount;
      pEhResCall->pEhResList = pEhResList; /* pointer back to parent */

      /* add node to linked list */
      //EHDBG("about to call ehAddHead()");
      //ehAddHead(&(pEhResTrk->EhList), &(pEhResCall->EhNode));
      //EHDBG("back from calling ehAddHead()");

   } EHENDCATCH;

   if (ehError())
   {
   }
   else
   {
      EHREM("--- resource updated ---");
   }
   if (pEhResList->pcResNam) EHREM1("pEhResList->pcResNam: %s",pEhResList->pcResNam);
   EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
   EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
   EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);
   if (pEhResCall) EHREM1("pEhResCall->pulArgs: %p",  pEhResCall->pulArgs);

   EHRETURN(pEhResList);
} /* update_old_node */


/***
reuse_old_node()
Reuse an old (stale) node in resource tree which is no longer in use.
***/

static
void
reuse_old_node A4(
   EHCALL *,     pEhCall,
   EHRESTRK *,   pEhResTrk,
   EHRESLIST *,  pEhResList,
   ulong **,     ppulArgs)
{
   EHRESCALL *volatile pEhResCall = NULL;

   EHCATCH(reuse_old_node) {

      /* it must be RELEASED */
      EHASSERT_EXITSEQ(pEhResList->EhResState == EH_RES_RELEASED);
      /* there must be at least one EHRESCALL rec */
      EHASSERT_EXITSEQ(pEhResList->szRCAlloc >= 1);
      /* there must be at least one EHRESCALL rec */
      EHASSERT_EXITSEQ(pEhResList->pEhResCall != NULL);

      EHREM("--- reusing old node ---");
      EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
      EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
      EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);

      /* update tracking info */
      pEhResList->EhResState = EH_RES_OBTAINED;
      pEhResList->EhResWhom = EH_RES_OBTAIN;
      pEhResList->uhNumObtn = 1; /* this is true for *this* resource */
      pEhResList->szRCInUse = 1;
      pEhResList->hasProblem = false;
      pEhResCall = pEhResList->pEhResCall;

      /* set values for instance of resource function call */
      pEhResCall->pulArgs = *ppulArgs;
      *ppulArgs = NULL;
      pEhResCall->pcFileO  = pEhCall->pcFile;
      pEhResCall->pcFuncO  = pEhCall->pcFunc;
      pEhResCall->iLineO   = pEhCall->iLine;
      pEhResCall->ulCountO = pEhCall->ulCount;
      pEhResCall->pcFileR = NULL;
      pEhResCall->pcFuncR = NULL;
      pEhResCall->iLineR = 0;
      pEhResCall->ulCountR = 0L;
      pEhResCall->pEhResList = pEhResList; /* pointer back to parent */

      /* add node to linked list */
      ehAddHead(&(pEhResTrk->EhList), &(pEhResCall->EhNode));

   } EHENDCATCH;

   if (ehError())
   {
      if (pEhResCall->pulArgs) /* revert */
      {
         *ppulArgs = pEhResCall->pulArgs;
         pEhResCall->pulArgs = NULL;
      }
      /* set values for instance of resource function call */
      pEhResCall->pcFileR  = pEhCall->pcFile;
      pEhResCall->pcFuncR  = pEhCall->pcFunc;
      pEhResCall->iLineR   = pEhCall->iLine;
      pEhResCall->ulCountR = pEhResCall->ulCountO;

      /* update tracking info */
      pEhResList->EhResState = EH_RES_RELEASED;
      pEhResList->szRCInUse = 0;
      pEhResList->uhNumObtn = 0;
   }
   else
   {
      EHREM("--- old node reused ---");
   }
   if (pEhResList->pcResNam) EHREM1("pEhResList->pcResNam: %s",pEhResList->pcResNam);
   EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
   EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
   EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);
   if (pEhResCall) EHREM1("pEhResCall->pulArgs: %p",  pEhResCall->pulArgs);

   EHLEAVE();
} /* reuse_old_node */


/***
ehResRel()
Track resource de-allocation.
***/

void
ehResRel A7V(
   const char *, pcFile,
   const char *, pcFunc,
   int,          iLine,
   EHRESTRK *,   pEhResTrk,   /* resource tracking header */
   int,          iType,
   pfv_ulv,      pfvRelease,  /* ptr func ret void */
   ulong,        ulFlags,     /* some of these help us find the handle */
   ...)
{
   /* used in Clean Up section */
   EHRESCALL *volatile pEhResCall = NULL;
   EHRESLIST *volatile pEhResList = NULL;
   volatile int iNumArgs = (int)(ulFlags & EH_RES_NUMARGS_MSK);
   ulong aulArgs[EH_RES_NUMARGS_MSK]; /* for error recovery */

   volatile ulong ulResHndl = 0L;

   EHCATCH(ehResRel) {

      /* not used in Clean Up section */
      va_list ap;
      EHRESLIST  ResList;
      int iArg;

      if (cEhResDmp)
      {
         fprintf(EHSTDERR, "ehResRel(%s, %s, %d, %d, 0x%lx, ..., ", 
               pcFile, pcFunc?pcFunc:"<func>", iLine, iType, ulFlags);
      }

      /*** create argument vector and fetch handle ***/

      if (iNumArgs > 0)
      {
         /* load array with arguments */
         VA_START(ap, ulFlags); 
         for (iArg = 0; iArg < iNumArgs; ++iArg)
         {
            aulArgs[iArg] = va_arg(ap, ulong);
         }
         va_end(ap);
      }

      /* pick off the resource handle from the argument list */
      EHASSERT_EXITSEQ( (ulFlags & EH_RES_HNDL_2ND) == 0 || (iNumArgs > 1) );
      if (iNumArgs > 0)
      {
         ulResHndl = aulArgs[ (ulFlags & EH_RES_HNDL_2ND) ? 1 : 0 ];
      }
      else /* iNumArgs == 0 */
      {
         EHASSERT_EXITSEQ(iNumArgs == 0);
         ulResHndl = EH_RES_VOID;
      }
      if (cEhResDmp)
      {
         fprintf(EHSTDERR, "0x%lx)\n", ulResHndl);
         fflush(EHSTDERR);
         ehAvlTreeDump(pEhResTrk->pEhAvlRes);
      }
      EHASSERT_EXITSEQ(ulResHndl != 0L);

      /*** update array ***/

      /* find current EHRESLIST entry for given pointer */
      ResList.ulResHndl = ulResHndl;
      ResList.iType = iType;
      if (cEhAvlDmp) ehAvlTreeDump(pEhResTrk->pEhAvlRes);
      if ((pEhResList = (EHRESLIST*)
            ehAvlSrch(pEhResTrk->pEhAvlRes, &ResList)) == NULL)
      {
         ulong ulCount = update_pass_count(pEhResTrk, iLine, pcFile);
         //EHERR1("EhCall.ulCount = %lu", EhCall.ulCount);
         ehIndent(EHSTDERR);
         fprintf(EHSTDERR, "*** invalid resource handle %lx ***\n",
               ResList.ulResHndl);
         ehIndent(EHSTDERR);
         fprintf(EHSTDERR, "in function: %s %s%s%u pass %lu\n",
               pcFile, 
               pcFunc?pcFunc:"",
               pcFunc?"() ":"",
               iLine,
               ulCount);
         fflush(EHSTDERR);
         EHTHROW(EH_LOSTRESOURCE);   /* ? lost pointer ? */
      }

      /*** update tracking info ***/

      EHREM("--- about to release a resource ---");
      EHREM1("resource handle is 0x%lx", ulResHndl);
      EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
      EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
      EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);

      if (pEhResList->EhResState == EH_RES_RELEASED &&
          pEhResList->szRCInUse <= 0) /* released too many times */
      {
         if (pEhResList->EhResWhom == EH_RES_RELEASE)
         {
            size_t   szRCAlloc;
            ulong ulCount = update_pass_count(pEhResTrk, iLine, pcFile);

            if (pEhResList->pcResNam)
            {
               ehIndent(EHSTDERR);
               fprintf(EHSTDERR, "resource: \"%s\"\n", pEhResList->pcResNam);
            }
            /***
               If the resource is already marked as released, it is likely that
               the szRCInUse counter is already zero.  Therefore, we should use
               the szRCAlloc counter for these error messages.
            ***/
            for (pEhResCall = pEhResList->pEhResCall, szRCAlloc = pEhResList->szRCAlloc;
                 szRCAlloc--; ++pEhResCall)
            {
               ehIndent(EHSTDERR);
               fprintf(EHSTDERR, "originally obtained: %s %s%s%u pass %lu\n", 
                     pEhResCall->pcFileO, 
                     pEhResCall->pcFuncO?pEhResCall->pcFuncO:"", 
                     pEhResCall->pcFuncO?"() ":"", 
                     pEhResCall->iLineO,
                     pEhResCall->ulCountO);
               if (pEhResCall->iLineR)
               {
                  ehIndent(EHSTDERR);
                  fprintf(EHSTDERR, "previously released: %s %s%s%u pass %lu\n", 
                        pEhResCall->pcFileR, 
                        pEhResCall->pcFuncR?pEhResCall->pcFuncR:"", 
                        pEhResCall->pcFuncR?"() ":"", 
                        pEhResCall->iLineR,
                        pEhResCall->ulCountR);
               }
            }
            ehIndent(EHSTDERR);
            fprintf(EHSTDERR, "released again at:   %s %s%s%u pass %lu\n", 
                  pcFile, pcFunc?pcFunc:"",pcFunc?"() ":"", iLine,
                  ulCount);
            fflush(EHSTDERR);
            EHTHROW(EH_RELEASEDTWICE);
         }
         break; /* someone else has released the resource, its ok */
      }

      /* test special case for RES_MEM */

      if ( (pEhResList->ulFlags & EH_RES_A1_EQU_A2) || (ulFlags & EH_RES_A1_EQU_A2) )
      {
         /* see if number allocated is the same as number freed */
         if (pEhResList->pEhResCall->pulArgs[1] != aulArgs[1])
         {
            ehIndent(EHSTDERR);
            fprintf(EHSTDERR, "at %s %s%s%u, %u was obtained\n", 
               pEhResList->pEhResCall->pcFileO, 
               pEhResList->pEhResCall->pcFuncO?pEhResList->pEhResCall->pcFuncO:"",
               pEhResList->pEhResCall->pcFuncO?"() ":"",
               pEhResList->pEhResCall->iLineO,  
               pEhResList->pEhResCall->pulArgs[1]);
            ehIndent(EHSTDERR);
            fprintf(EHSTDERR, "at %s %s%s%u, %u was released!\n", 
               pcFile, pcFunc?pcFunc:"",pcFunc?"() ":"", iLine, aulArgs[1]);
            fflush(EHSTDERR);
            EHTHROW(EH_BADARG);
         }
      }

      /*** should be safe to release resource now ***/

      if (pfvRelease) /* NULL is allowed */
      {
         /* switch on # args and use appropriate function call */
         switch(iNumArgs)
         {
         case 0:
            (*pfvRelease)(0L); /* dummy argument */
            break;
         case 1:
            (*pfvRelease)(aulArgs[0]);
            break;
         case 2:
            (*pfvRelease)(aulArgs[0], aulArgs[1]);
            break;
         case 3:
            (*pfvRelease)(aulArgs[0], aulArgs[1], aulArgs[2]);
            break;
         default: EHTHROW(EH_SWITCH);
         }
      }
      ulResHndl = 0L;

      /*** no more possible failures, now muck with data structures ***/

      ++pEhResTrk->ulResReleased;    /* how many resources were released */
      if (--(pEhResList->szRCInUse) == 0)
      {
         pEhResList->EhResState = EH_RES_RELEASED;
      }
      pEhResList->pfvActRel = pfvRelease; /* remember actual releasing function */
      pEhResList->EhResWhom = EH_RES_RELEASE;
      if (pEhResList->pEhResCall) /* this can be zero if errors occur */
      {
         pEhResCall = pEhResList->pEhResCall + pEhResList->szRCInUse;
         pEhResCall->pcFileR  = pcFile;
         pEhResCall->pcFuncR  = pcFunc;
         pEhResCall->iLineR   = iLine;
         pEhResCall->ulCountR = update_pass_count(pEhResTrk, iLine, pcFile);
         ehRemove(&(pEhResCall->EhNode));   /* remove node from linked list */
      }

   } EHENDCATCH;

   /* free stuff only within Clean Up section */
   if (!ehError())
   {
      if (pEhResList->szRCInUse == 0)
      {
         if (pEhResList->pcResNam) 
         {
            free( (char *)(((EHRESLIST *)pEhResList)->pcResNam) );
            pEhResList->pcResNam = NULL;
         }
      }
      if (pEhResCall)
         if (pEhResCall->pulArgs)
            EHFREE(pEhResCall->pulArgs);

      EHREM("--- resource released ---");
   }

   if (cEhAvlDmp) ehAvlTreeDump(pEhResTrk->pEhAvlRes);
   if (cEhResDmp) ehResTreeDump(pEhResTrk->pEhAvlRes);

   /* if a user-requested interrupt, let's try one more time */
   /* otherwise the resource won't get released */
   if (ehError() && ehCtrlC() && ulResHndl && pfvRelease)
   {
      if (pEhResList) /* otherwise use the ones passed in (gasp) */
         ulFlags = pEhResList->ulFlags;
      switch(iNumArgs)
      {
      case 0:
         ehResRel(pcFile,pcFunc,iLine,pEhResTrk,
               iType, pfvRelease, ulFlags);
         break;
      case 1:
         ehResRel(pcFile,pcFunc,iLine,pEhResTrk,
               iType, pfvRelease, ulFlags, aulArgs[0]);
         break;
      case 2:
         ehResRel(pcFile,pcFunc,iLine,pEhResTrk,
               iType, pfvRelease, ulFlags, aulArgs[0], aulArgs[1]);
         break;
      case 3:
         ehResRel(pcFile,pcFunc,iLine,pEhResTrk,
               iType, pfvRelease, ulFlags, aulArgs[0], aulArgs[1], aulArgs[2]);
         break;
      }
   }

   if (pEhResList)
   {
      if (pEhResList->pcResNam)EHREM1("pEhResList->pcResNam: %s",pEhResList->pcResNam);
      EHREM1("ResCalls allocated: %d", pEhResList->szRCAlloc);
      EHREM1("ResCalls bein used: %d", pEhResList->szRCInUse);
      EHREM1("ResCalls obtained:  %d", pEhResList->uhNumObtn);
      if (pEhResCall) EHREM1("pEhResCall->pulArgs: %p",  pEhResCall->pulArgs);
   }

   EHLEAVE();
} /* ehResRel */


/***

ehResReObt()
Re-obtain resource.
First we attempt to release the existing resource,
the re-obtain the new resource.

Error Handling:
   (1) find_old_node() fails
       -- 
       [a] set ulResHndlNew to NULL
       [b] mark orignal node as RELEASED
   (2) compare_handles() fails
       -- 
       [a] set ulResHndlNew to NULL
       [b] mark orignal node as RELEASED
       [c] fix existing RESLIST node, if one exists
       [d] release the resource (?)

***/

bool
ehResReObt A9V(
   const char*, pcFile,
   const char*, pcFunc,
   int,         iLine,
   EHRESTRK*,   pEhResTrk,    /* resource tracking header */
   int,         iType,
   const char*, pcResName,    /* text name for resource */
   ulong,       ulResHndlOld, /* old resource handle */
   pfv_ulv,     pfvRelease,   /* funct to release resource */
   ulong,       ulFlags,      /* argv count and other hints */
   ...)
{
   /* used in Clean Up section */
   EHRESLIST *volatile pEhResListOld = NULL;
   EHRESLIST *volatile pEhResList = NULL;
   ulong     *volatile pulArgs = NULL;
   volatile int iNumArgs = (int)(ulFlags & EH_RES_NUMARGS_MSK);
   volatile ulong ulResHndlNew = 0L;
   ulong aulArgs[EH_RES_NUMARGS_MSK]; /* for error recovery */

   EHCATCH(ehResReObt) {

      /* not used in Clean Up section */
      va_list ap;
      EHRESLIST  ResList;
      int iArg;
      EHCALL EhCall;

      if (cEhResDmp)
      {
         fprintf(EHSTDERR, "ehResReObt(%s, %s, %d, %d, 0x%lx,..., 0x%lx->",
               pcFile, pcFunc?pcFunc:"<func>", iLine, iType, ulFlags, 
               ulResHndlOld);
         fflush(EHSTDERR);
      }

      /***
         If realloc moved its block, there is an old block
         which is no longer allocated but which the error
         handler still thinks is allocated.  If we allocate
         any memory, we may inadvertantly obtain memory in
         the same region.  This causes the bounds checking
         routines to complain.  The first order of business
         is, therefore, to mark the old block as RELEASED.
      ***/

      /* create argument vector */
      if (iNumArgs > 0)
      {
         /* load array with arguments */
         VA_START(ap, ulFlags); 
         for (iArg = 0; iArg < iNumArgs; ++iArg)
         {
            aulArgs[iArg] = va_arg(ap, ulong);
         }
         va_end(ap);
      }

      /* set ResList.ulResHndl and ResList.iType */
      if (ulFlags & EH_RES_VOID)
      {
         EHASSERT_EXITSEQ(iNumArgs == 0);
         ulResHndlNew = EH_RES_VOID;
      }
      else if (ulFlags & EH_RES_HNDL_2ND)
      {
         EHASSERT_EXITSEQ(iNumArgs > 1);
         ulResHndlNew = aulArgs[1];
      }
      else
      {
         ulResHndlNew = aulArgs[0];
      }

      if (cEhResDmp)
      {
         fprintf(EHSTDERR, "0x%lx)\n", ulResHndlNew);
         fflush(EHSTDERR);
         ehAvlTreeDump(pEhResTrk->pEhAvlRes);
      }

      //printf("ehResReObt: ulResHndlNew = %lx\n", ulResHndlNew);
      EHASSERT_EXITSEQ(ulResHndlNew != 0L);
      EHDBG1("ulResHndlOld = %lx", ulResHndlOld);
      EHDBG1("ulResHndlNew = %lx", ulResHndlNew);

      /*** locate old resource node ***/

      if (ulResHndlOld != 0) /* old handle might be null for realloc() */
      {
         EHRESLIST ResListOld;
         ResListOld.ulResHndl = ulResHndlOld;
         ResListOld.iType = iType;
         pEhResListOld = find_old_node(pEhResTrk, &ResListOld, ulFlags);
         if (pEhResListOld == NULL)
         {
            ulong ulCount = update_pass_count(pEhResTrk, iLine, pcFile);
            ehIndent(EHSTDERR);
            fprintf(EHSTDERR, "*** invalid resource handle %lx ***\n",
                  ResListOld.ulResHndl);
            ehIndent(EHSTDERR);
            fprintf(EHSTDERR, "in function: %s %s%s%u pass %lu\n",
                  pcFile,
                  pcFunc ? pcFunc : "",
                  pcFunc ? "() " : "",
                  iLine,
                  ulCount);
            fflush(EHSTDERR);
            EHTHROW(EH_LOSTHANDLE);
         }
         if (ulResHndlOld != ulResHndlNew)
         {
            /* supress the bounds checker */
            pEhResListOld->EhResState = EH_RES_RELEASED; /* temporarily */
            pEhResListOld->EhResWhom  = EH_RES_REOBTAIN;
         }
      }

      EhCall.pcFile = pcFile;
      EhCall.pcFunc = pcFunc;
      EhCall.iLine  = iLine;
      EhCall.ulCount = update_pass_count(pEhResTrk, iLine, pcFile);
      //EHERR1("EhCall.ulCount = %lu", EhCall.ulCount);

      /*** parse variable arguments that will be used by releasing routine ***/

      /* create argument vector */
      if (iNumArgs > 0)
      {
         /* create temp argument array of longs */
         if ((pulArgs = (ulong*)calloc(sizeof(ulong), iNumArgs)) == NULL)
               EHTHROW(EH_CALLOC);

         /* load array with arguments */
         memcpy(pulArgs, aulArgs, sizeof(ulong) * iNumArgs);
      }

      /*** compare the old and new resource handles ***/

      /* initialize ResList */
      memset(&ResList, 0, sizeof(ResList));
      ResList.ulFlags = ulFlags;
      ResList.pfvDefRel = pfvRelease;
      ResList.EhResState = EH_RES_OBTAINED;
      ResList.EhResWhom  = EH_RES_NONE;

      /* compare handles and update tree */
      ResList.ulResHndl = ulResHndlNew;
      ResList.iType = iType;
      pEhResList = compare_handles(&EhCall, pEhResTrk,
            pEhResListOld, &ResList, (ulong**)&pulArgs);

      /*** copy resource name (it may be different) ***/

      EHASSERT_EXITSEQ(pcResName != NULL);
      if (pEhResList->pcResNam) EHFREE(pEhResList->pcResNam);
      if ((pEhResList->pcResNam = (char*)calloc(strlen(pcResName)+1,1)) == NULL)
         EHTHROW(EH_CALLOC);
      strcpy(pEhResList->pcResNam, pcResName);
      pEhResList->pfvActRel = pfvRelease; /* remember actual releasing function */
      pEhResList->EhResWhom = EH_RES_REOBTAIN;

      /*** no more possible failures, now muck with data structures ***/

      EHDBGX(ehResTreeDump(pEhResTrk->pEhAvlRes));

      if (ulResHndlOld == 0)
      {
         ++pEhResTrk->ulResObtained; /* obtained a new resource */
      }
      else if (ulResHndlOld != ulResHndlNew)
      {
         ++pEhResTrk->ulResReleased; /* we released a resource... */
         ++pEhResTrk->ulResObtained; /* and obtained one. */
      }

   } EHENDCATCH;

   EHDBG("in Clean Up section");
   /* free stuff only within Clean Up section */
   if (!ehError())
   {
      EHREM("--- resource released ---");
   }

   if (cEhAvlDmp) ehAvlTreeDump(pEhResTrk->pEhAvlRes);
   if (cEhResDmp) ehResTreeDump(pEhResTrk->pEhAvlRes);

   /* pulArgs was allocated, free it */
   if (pulArgs) free(pulArgs);

   EHDBG("about to call EHRETURN(true)");
   EHRETURN(true);
} /* ehResReObt */


/***

Find the RESLIST associated with
the handle for a currently active
(obtained) resource.

***/

static
EHRESLIST *
find_old_node A3(
   EHRESTRK *,  pEhResTrk,  /* resource tracking anchor */
   EHRESLIST *, pEhResList, /* must have handle and type */
   ulong,       ulFlags)    /* some of these help us find the handle */
{
   EHRESLIST *volatile pEhResListOld = NULL;

   EHTRACE(find_old_node);

   /*** allow null pointer only if flags allow ***/

   if (pEhResList->ulResHndl == 0L)
   {
      if (ulFlags & EH_RES_ALLOW_NULL)
      {
         EHDBG("can't look for null resource handle");
         EHRETURN((EHRESLIST*)NULL); /* OK, let caller know it was a NULL */
      }
      else
      {
         EHERR("null handles not allowed when re-acquiring this resource");
         EHTHROW_RETURN(EH_NULLHANDLE, (EHRESLIST*)NULL); /* error */
      }
   }

   /*** search for handle in resource tree ***/

   pEhResListOld =
         (EHRESLIST*)ehAvlSrch(pEhResTrk->pEhAvlRes, (const void*)pEhResList);
   if (pEhResListOld == NULL)
   {
      EHDBG2("Couldn't find handle %p type %d in resource tree",
            pEhResList->ulResHndl, pEhResList->iType);
      EHRETURN((EHRESLIST*)NULL);
   }

   /*** make sure resource is currently active/obtained ***/

   EHDBGX(ehResTreeDump(pEhResTrk->pEhAvlRes));
   EHASSERT_RETURN(pEhResListOld->ulResHndl == pEhResList->ulResHndl,
         (EHRESLIST*)NULL);
   EHASSERT_RETURN(pEhResListOld->EhResState == EH_RES_OBTAINED,
         (EHRESLIST*)NULL);

   EHRETURN(pEhResListOld);
} /* find_old_node */


/***

This function compares the old resource handle
with the new resource handle when a resource 
has been reobtained.  They can be:
   1) different
      a) new handle is already in resource tree
      b) new handle is not in resource tree
   2) the same

Error Handling:
   (1) ehAvlSrch() fails
       nothing
   (2) ASSERT_EXITSEQ(EH_RES_RELEASED)
       nothing
   (3) ??? ehInSort() fails
       nothing
   (4) ASSERT_EXITSEQ(EH_RES_OBTAINED)
       nothing

***/

static
EHRESLIST*
compare_handles A5(
   EHCALL *,    pEhCall,
   EHRESTRK*,   pEhResTrk,      /* resource tracking anchor */
   EHRESLIST*,  pEhResListOld,  /* NULL means there was no original block */
   EHRESLIST*,  pEhResListTmp,  /* contains handle, type and flags */
   ulong **,    ppulArgsNew)    /* new release arguments */
{
   EHRESLIST*volatile pEhResListNew = NULL;
   EHRESCALL*volatile pEhResCall = NULL;
   volatile bool removed = false;

   EHCATCH(compare_handles) {

      /*** check to see if handles are different ***/

      if (pEhResListOld == NULL ||
            pEhResListOld->ulResHndl != pEhResListTmp->ulResHndl)
      {
         /*** 1) handles are different ***/

         if (pEhResListOld != NULL) /* make sure there was an old block */
         {
            /* mark old resource node as released */
            pEhResListOld->EhResState = EH_RES_RELEASED;
            pEhResListOld->EhResWhom  = EH_RES_REOBTAIN;
            if (pEhResListOld->pcResNam) EHFREE(pEhResListOld->pcResNam);
            /* pEhResListOld->pfvActRel = pfvRelease; */

            /* update function call info */
            EHASSERT_EXITSEQ(pEhResListOld->szRCInUse == 1);
            pEhResCall = pEhResListOld->pEhResCall;
            pEhResCall->pcFileR    = pEhCall->pcFile;
            pEhResCall->pcFuncR    = pEhCall->pcFunc;
            pEhResCall->iLineR     = pEhCall->iLine;
            pEhResCall->ulCountR   = pEhCall->ulCount;
            if (pEhResCall->pulArgs) EHFREE(pEhResCall->pulArgs);
            removed = ehRemove(&(pEhResCall->EhNode));   /* remove node from linked list */
         }

         /*** does new handle already have an (old) entry in the resource tree ***/

         pEhResListNew = (EHRESLIST*)
               ehAvlSrch(pEhResTrk->pEhAvlRes, pEhResListTmp);

         if (pEhResListNew != NULL)
         {
            /*** 1a) new handle was already in resource tree, reuse old node ***/

            /* existing node MUST contain a released resource */
            EHASSERT_EXITSEQ(pEhResListNew->EhResState == EH_RES_RELEASED);

            /* (save for recovery ?) */

            /* revise node */
            reuse_old_node(pEhCall, pEhResTrk, pEhResListNew, ppulArgsNew);
         }
         else
         {
            /*** 1b) new handle not already in tree, insert new node ***/

            /* insert node into tree */
            pEhResListNew = insert_new_node(pEhCall, pEhResTrk,
                  pEhResListTmp, ppulArgsNew);
            EHDBG("returned from insert_new_node()");
         }
      }
      else /* pEhResListOld != NULL && pEhResListOld->ulHndl == ulResHndlNew */
      {
         /*** 2) new and old handles are the same, update old node ***/

         /* affirm that this handle is, in fact, in use */
         EHASSERT_EXITSEQ(pEhResListOld->EhResState == EH_RES_OBTAINED);

         /* (save for recovery ?) */

         pEhResListNew = update_old_node(pEhCall, pEhResTrk,
               pEhResListOld, ppulArgsNew);
      }

   } EHENDCATCH;

   EHDBG("in Clean Up section");
   if (ehError())
   {

      /* ??? if inserted ??? */
         /* ??? remove ??? */
         /* ??? mark as released ??? */

      if (pEhResListNew) /* existing node was modified;  restore */
      {
         EHRESCALL *pEhResCall;

         pEhResListNew->EhResState = EH_RES_RELEASED;
         pEhResListNew->EhResWhom  = EH_RES_REOBTAIN;
         pEhResListNew->hasProblem = false;

         pEhResCall = pEhResListNew->pEhResCall;
         pEhResCall->pcFileO    = pEhCall->pcFile;
         pEhResCall->pcFuncO    = pEhCall->pcFunc;
         pEhResCall->iLineO     = pEhCall->iLine;
         pEhResCall->ulCountO   = pEhCall->ulCount;
         pEhResCall->pcFileR    = NULL;
         pEhResCall->pcFuncR    = NULL;
         pEhResCall->iLineR     = 0;
         pEhResCall->ulCountR   = 0;
         pEhResCall->pcFileR    = pEhCall->pcFile;
         pEhResCall->pcFuncR    = pEhCall->pcFunc;
         pEhResCall->iLineR     = pEhCall->iLine;
         pEhResCall->ulCountR   = pEhCall->ulCount;
      }
      if (removed && pEhResCall)
      {
         /* since we can't put it back exactly, at least put it back */
         ehAddHead(&(pEhResTrk->EhList), &(pEhResCall->EhNode));
      }
   }
   else
   {
      /* ??? free the old resource name buffers ??? */
      /* ??? if (newSave) free(ResListNewSave.pcResource); ??? */
      /* ??? if (oldSave) free(ResListOldSave.pcResource); ??? */
   }
   EHDBG("about to call EHLEAVE()");
   EHRETURN(pEhResListNew);
} /* compare_handles */


/***
ehVerify()
Verify that the specified resource
exists in the resource tree and
is of the specified type and
is currently obtained.
***/

bool
ehResVerify A3(
   EHRESTRK *, pEhResTrk,   /* resource tracking header */
   ulong,      ulResHndl,
   int,        iType)
{
   EHRESLIST *pEhResList = NULL;
   EHRESLIST ResList;

   EHTRACE(ehResVerify);

   /* find current EHRESLIST entry for given pointer */
   ResList.ulResHndl = ulResHndl;
   ResList.iType = iType;
   if ((pEhResList = (EHRESLIST*)
         ehAvlSrch(pEhResTrk->pEhAvlRes, &ResList)) == NULL)
   {
      EHRETURN(false);
   }

   if (pEhResList->EhResState != EH_RES_OBTAINED)
   {
      EHRETURN(false);
   }

   EHRETURN(true);
} /* ehResVerify */

/***
ehResRelBy()
Return a pointer to the function
which last released this resource,
NULL if none.
***/

pfv_ulv
ehResRelBy A3(
   EHRESTRK *, pEhResTrk,   /* resource tracking header */
   ulong,      ulResHndl,
   int,        iType)
{
   EHRESLIST *pEhResList = NULL;
   EHRESLIST ResList;

   EHTRACE(ehResRelBy);

   /* find current EHRESLIST entry for given pointer */
   ResList.ulResHndl = ulResHndl;
   ResList.iType = iType;
   if ((pEhResList = (EHRESLIST*)
         ehAvlSrch(pEhResTrk->pEhAvlRes, &ResList)) == NULL)
   {
      EHRETURN((pfv_ulv)NULL);
   }

   EHRETURN(pEhResList->pfvActRel);
} /* ehResRelBy */


/***
ehResSetRelBy()
Set the pointer in the EHRESLIST
structure which tracks the function
which last released this resource.
***/

void
ehResSetRelBy A4(
   EHRESTRK *, pEhResTrk,   /* resource tracking header */
   ulong,      ulResHndl,
   int,        iType,
   pfv_ulv,    pfvRelease)
{
   EHRESLIST *pEhResList = NULL;
   EHRESLIST ResList;

   EHTRACE(ehResSetRelBy);

   /* find current EHRESLIST entry for given pointer */
   ResList.ulResHndl = ulResHndl;
   ResList.iType = iType;
   if ((pEhResList = (EHRESLIST*)
         ehAvlSrch(pEhResTrk->pEhAvlRes, &ResList)) == NULL)
   {
      EHTHROW_LEAVE(EH_AVLSRCH);
   }

   pEhResList->pfvActRel = pfvRelease;

   EHLEAVE();
} /* ehResSetRelBy */


/***
This function traverses the
EHRESFUN tree to find the
entry for the given file and line.
If found, the count value is 
incremented and returned.
If not found, a new structure
is allocated and the count
value is set to one.
***/

static
ulong
update_pass_count A3(
   EHRESTRK *,   pEhResTrk,
   int,          iLine,
   const char *, pcFile)
{
   EHRESFUNC *volatile pEhResFuncNew = NULL;
   EHRESFUNC *pEhResFuncOld = NULL;
   ulong ulPassCount = 0;
   EHRESFUNC EhResFunc;

   EHCATCH(update_pass_count) {

      /*** search the tree ***/

      EhResFunc.iLine = iLine;
      EhResFunc.pcFile = pcFile;
      pEhResFuncOld =  (EHRESFUNC *)
                   ehAvlSrch(pEhResTrk->pEhAvlFun, (const void*)&EhResFunc);

      if (pEhResFuncOld == NULL)
      {
         /*** add new entry ***/

         if ((pEhResFuncNew = NEW(EHRESFUNC)) == NULL)
            EHTHROW_RETURN(EH_NEWMACRO, 0L);
         *pEhResFuncNew = EhResFunc;
         ulPassCount = pEhResFuncNew->ulCount = 1L;

         /* insert new ResFunc into tree */
         ehAvlAdd(
               pEhResTrk->pEhAvlFun,
               (void *)pEhResFuncNew,
               (pfi_pvpv)NULL);
      }
      else
      {
         /*** increment old entry ***/

         ulPassCount = ++(pEhResFuncOld->ulCount);
      }

   } EHENDCATCH;

   if (ehError())
   {
      if (pEhResFuncNew)
      {
         free(pEhResFuncNew);
         ulPassCount = 0;
      }
   }
   EHRETURN(ulPassCount);
}
